# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.combinat
r"""
Homsets of finitely presented graded modules

TESTS::

    sage: from sage.modules.fp_graded.module import FPModule
    sage: A = SteenrodAlgebra(2, profile=(3,2,1))
    sage: F = A.free_graded_module([1,3], names='c')
    sage: L = FPModule(A, [2,3], [[Sq(2),Sq(1)], [0,Sq(2)]], names='h')
    sage: homset = Hom(F, L); homset
    Set of Morphisms from Free graded left module on 2 generators ...
    sage: homset.an_element()
    Module morphism:
      From: Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
      To:   Finitely presented left module on 2 generators and 2 relations over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
      Defn: c[1] |--> 0
            c[3] |--> Sq(1)*h[2]
    sage: f = homset([L((A.Sq(1), 1)), L((0, A.Sq(2)))]); f
    Module morphism:
      From: Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
      To:   Finitely presented left module on 2 generators and 2 relations over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
      Defn: c[1] |--> Sq(1)*h[2] + h[3]
            c[3] |--> Sq(2)*h[3]
    sage: f.kernel_inclusion()
    Module morphism:
      From: Finitely presented left module on 2 generators and 1 relation over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
      To:   Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
      Defn: g[3] |--> c[3]
            g[4] |--> Sq(0,1)*c[1]

AUTHORS:

- Robert R. Bruner, Michael J. Catanzaro (2012): Initial version.
- Sverre Lunoee--Nielsen and Koen van Woerden (2019-11-29): Updated the
  original software to Sage version 8.9.
- Sverre Lunoee--Nielsen (2020-07-01): Refactored the code and added
  new documentation and tests.
"""

# *****************************************************************************
#       Copyright (C) 2011 Robert R. Bruner <rrb@math.wayne.edu> and
#                          Michael J. Catanzaro <mike@math.wayne.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.homset import Homset


class FPModuleHomspace(Homset):
    # FPModuleMorphism imports FPModuleHomspace, so this import should
    # not happen at the top level.
    from .morphism import FPModuleMorphism as Element

    def _element_constructor_(self, values):
        r"""
        Construct a morphism contained in ``self``.

        This function is not part of the public API, but is used by
        :meth:`Hom` method to create morphisms.

        INPUT:

        - ``values`` -- an iterable of FPElements of the codomain

        OUTPUT:

        A module homomorphism in this homspace sending the generators
        of the domain module to the given values.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))
            sage: F = FPModule(A2, [1,3])
            sage: L = FPModule(A2, [2,3], [[Sq(2),Sq(1)], [0,Sq(2)]])

            sage: homset = Hom(F, L)
            sage: v1 = L([A2.Sq(1), 1])
            sage: v2 = L([0, A2.Sq(2)])
            sage: f = homset._element_constructor_([v1, v2])

        Rather than calling ``_element_constructor_`` explicitly, one
        should call it implicitly using the syntax::

            sage: f = homset([v1, v2]); f
            Module morphism:
              From: Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
              To:   Finitely presented left module on 2 generators and 2 relations over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
              Defn: g[1] |--> Sq(1)*g[2] + g[3]
                    g[3] |--> Sq(2)*g[3]

        One can construct a homomorphism from another homomorphism::

            sage: g = homset(f)
            sage: f == g
            True

        And there is a convenient way of making the trivial homomorphism::

            sage: homset(0) == 0
            True

        TESTS::

            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))
            sage: F = A2.free_graded_module((1,3))
            sage: L = A2.free_graded_module((2,5))
            sage: H = Hom(F, L)

            sage: values = (A2.Sq(4)*L.generator(0), A2.Sq(3)*L.generator(1))
            sage: f = H(values); f
            Module morphism:
              From: Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
              To:   Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
              Defn: g[1] |--> Sq(4)*g[2]
                    g[3] |--> Sq(3)*g[5]

            sage: H(0)
            Module morphism:
              From: Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
              To:   Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
              Defn: g[1] |--> 0
                    g[3] |--> 0
        """
        if isinstance(values, self.element_class):
            return values
        elif values == 0 or all(v.is_zero() for v in values):
            return self.zero()
        else:
            return self.element_class(self, values)

    def an_element(self, n=0):
        r"""
        Create a homomorphism belonging to ``self``.

        INPUT:

        - ``n`` -- (default: 0) an integer degree

        OUTPUT: a module homomorphism of degree ``n``

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: HZ = FPModule(A, [0], relations=[[Sq(1)]])

            sage: Hom(HZ, HZ).an_element(3)
            Module endomorphism of Finitely presented left module on 1 generator and 1 relation over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> Sq(0,1)*g[0]

        TESTS::

            sage: K = FPModule(A, [0, 0], [[Sq(2), 0]]) # Using a zero coefficient in the relations.
            sage: Hom(K, K).an_element(4)
            Module endomorphism of Finitely presented left module on 2 generators and 1 relation over mod 2 Steenrod algebra, milnor basis
              Defn: g[0, 0] |--> 0
                    g[0, 1] |--> Sq(4)*g[0, 0]

            sage: K = FPModule(A, [0, 0], [[Sq(2), 0], [0,0], [Sq(4), Sq(2)*Sq(2)]])
            sage: Hom(K, K).an_element(n=3)
            Module endomorphism of Finitely presented left module on 2 generators and 2 relations over mod 2 Steenrod algebra, milnor basis
              Defn: g[0, 0] |--> 0
                    g[0, 1] |--> Sq(0,1)*g[0, 0]

        An example involving free modules::

            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))
            sage: F = A2.free_graded_module((1,3))
            sage: L = A2.free_graded_module((1,2))
            sage: H = Hom(F, L)
            sage: H.an_element()
            Module morphism:
              From: Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
              To:   Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
              Defn: g[1] |--> g[1]
                    g[3] |--> 0
        """
        return self._basis_elements(n, basis=False)

    def basis_elements(self, n):
        r"""
        Return a basis for the free module of degree ``n`` morphisms.

        INPUT:

        - ``n`` -- integer degree

        OUTPUT: a basis for the set of all module homomorphisms of degree ``n``

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: Hko = FPModule(A, [0], relations=[[Sq(2)], [Sq(1)]])

            sage: Hom(Hko, Hko).basis_elements(21)
            [Module endomorphism of Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
               Defn: g[0] |--> (Sq(0,0,3)+Sq(0,2,0,1))*g[0],
             Module endomorphism of Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
               Defn: g[0] |--> Sq(8,2,1)*g[0]]
        """
        return self._basis_elements(n, basis=True)

    def zero(self):
        r"""
        Create the trivial homomorphism in ``self``.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))
            sage: F = FPModule(A2, [1,3])
            sage: L = FPModule(A2, [2,3], [[Sq(2),Sq(1)], [0,Sq(2)]])

            sage: z = Hom(F, L).zero()
            sage: z(F.an_element(5))
            0
            sage: z(F.an_element(23))
            0

        Example with free modules::

            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))
            sage: F = A2.free_graded_module((1,3))
            sage: L = A2.free_graded_module((2,3))
            sage: H = Hom(F, L)
            sage: H.zero()
            Module morphism:
              From: Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
              To:   Free graded left module on 2 generators over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
              Defn: g[1] |--> 0
                    g[3] |--> 0
        """
        ngens = len(self.domain().generator_degrees())
        return self.element_class(self, [self.codomain().zero()] * ngens)

    def identity(self):
        r"""
        Return the identity homomorphism.

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))
            sage: L = FPModule(A2, [2,3], [[Sq(2),Sq(1)], [0,Sq(2)]])

            sage: one = Hom(L, L).identity(); one
            Module endomorphism of Finitely presented left module on 2 generators and 2 relations over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
              Defn: g[2] |--> g[2]
                    g[3] |--> g[3]

            sage: e = L.an_element(5)
            sage: e == one(e)
            True

        It is an error to call this function when the homset is not a
        set of endomorphisms::

            sage: F = FPModule(A2, [1,3])
            sage: Hom(F,L).identity()
            Traceback (most recent call last):
            ...
            TypeError: this homspace does not consist of endomorphisms

        An example with free graded modules::

            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))
            sage: L = A2.free_graded_module((2,3))
            sage: H = Hom(L, L)
            sage: H.identity()
            Module endomorphism of Free graded left module on 2 generators
             over sub-Hopf algebra of mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]
              Defn: g[2] |--> g[2]
                    g[3] |--> g[3]

        TESTS::

            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))
            sage: L = A2.free_graded_module((2,3))
            sage: F = A2.free_graded_module((1,3))
            sage: H = Hom(F, L)
            sage: H.identity()
            Traceback (most recent call last):
            ...
            TypeError: this homspace does not consist of endomorphisms
        """
        if self.is_endomorphism_set():
            return self.element_class(self, self.codomain().generators())
        else:
            raise TypeError('this homspace does not consist of endomorphisms')

    def _basis_elements(self, n, basis):
        r"""
        Return a basis for the free module of degree ``n`` homomorphisms.

        This function is private for use by :meth:`basis_elements` and
        :meth:`an_element`.

        INPUT:

        - ``n`` -- integer degree
        - ``basis`` -- boolean; decide if a basis should be returned or just
          a single homomorphism

        OUTPUT:

        A basis for the set of all module homomorphisms of degree ``n``
        if ``basis`` is ``True``. Otherwise a single element is returned.
        In the latter case, this homomorphism is non-trivial if the vector
        space of all homomorphisms is non-trivial.

        TESTS::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A = SteenrodAlgebra(2)
            sage: Hko = FPModule(A, [0], relations=[[Sq(2)], [Sq(1)]])
            sage: Hom(Hko, Hko)._basis_elements(21, basis=True)
            [Module endomorphism of Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
               Defn: g[0] |--> (Sq(0,0,3)+Sq(0,2,0,1))*g[0],
             Module endomorphism of Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
               Defn: g[0] |--> Sq(8,2,1)*g[0]]

            sage: Hom(Hko, Hko)._basis_elements(21, basis=False)
            Module endomorphism of Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> (Sq(0,0,3)+Sq(0,2,0,1))*g[0]

            sage: F = FPModule(A, [0])
            sage: Hom(F, Hko)._basis_elements(21, basis=False)
            Module morphism:
              From: Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              To:   Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> Sq(0,2,0,1)*g[0]

            sage: Hom(F, Hko)._basis_elements(21, basis=False)
            Module morphism:
              From: Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              To:   Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> Sq(0,2,0,1)*g[0]

        Test corner cases involving trivial modules::

            sage: F = FPModule(A, [0]) # A module without relations.
            sage: Z0 = FPModule(A, []) # A trivial module.
            sage: Z1 = FPModule(A, [0], [[1]]) # A trivial module with a redundant generator and relation.

            sage: Hom(FPModule(A, [-1]), F)._basis_elements(0, basis=True)
            []
            sage: Hom(FPModule(A, [-1]), F)._basis_elements(0, basis=False)
            Module morphism:
              From: Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              To:   Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              Defn: g[-1] |--> 0

            sage: from itertools import product
            sage: for D,C in product([(F, 'Free'), (Hko, 'Hko'), (Z0, 'Trivial'), (Z1, 'Trivial with redundant generator')], repeat=2):
            ....:     print('Hom(%s, %s):' % (D[1], C[1]))
            ....:     print('  basis==False:\n  %s' % Hom(D[0], C[0])._basis_elements(n=7, basis=False))
            ....:     print('  basis==True:\n  %s' % Hom(D[0], C[0])._basis_elements(n=7, basis=True))
            Hom(Free, Free):
              basis==False:
              Module endomorphism of Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> Sq(0,0,1)*g[0]
              basis==True:
              [Module endomorphism of Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> Sq(0,0,1)*g[0], Module endomorphism of Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> Sq(1,2)*g[0], Module endomorphism of Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> Sq(4,1)*g[0], Module endomorphism of Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> Sq(7)*g[0]]
            Hom(Free, Hko):
              basis==False:
              Module morphism:
              From: Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              To:   Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> Sq(0,0,1)*g[0]
              basis==True:
              [Module morphism:
              From: Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              To:   Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> Sq(0,0,1)*g[0]]
            Hom(Free, Trivial):
              basis==False:
              Module morphism:
              From: Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              To:   Free graded left module on 0 generators over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> 0
              basis==True:
              []
            Hom(Free, Trivial with redundant generator):
              basis==False:
              Module morphism:
              From: Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              To:   Finitely presented left module on 1 generator and 1 relation over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> 0
              basis==True:
              []
            Hom(Hko, Free):
              basis==False:
              Module morphism:
              From: Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
              To:   Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> 0
              basis==True:
              []
            Hom(Hko, Hko):
              basis==False:
              Module endomorphism of Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> Sq(0,0,1)*g[0]
              basis==True:
              [Module endomorphism of Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> Sq(0,0,1)*g[0]]
            Hom(Hko, Trivial):
              basis==False:
              Module morphism:
              From: Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
              To:   Free graded left module on 0 generators over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> 0
              basis==True:
              []
            Hom(Hko, Trivial with redundant generator):
              basis==False:
              Module morphism:
              From: Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
              To:   Finitely presented left module on 1 generator and 1 relation over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> 0
              basis==True:
              []
            Hom(Trivial, Free):
              basis==False:
              Module morphism:
              From: Free graded left module on 0 generators over mod 2 Steenrod algebra, milnor basis
              To:   Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              basis==True:
              []
            Hom(Trivial, Hko):
              basis==False:
              Module morphism:
              From: Free graded left module on 0 generators over mod 2 Steenrod algebra, milnor basis
              To:   Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
              basis==True:
              []
            Hom(Trivial, Trivial):
              basis==False:
              Module endomorphism of Free graded left module on 0 generators over mod 2 Steenrod algebra, milnor basis
              basis==True:
              []
            Hom(Trivial, Trivial with redundant generator):
              basis==False:
              Module morphism:
              From: Free graded left module on 0 generators over mod 2 Steenrod algebra, milnor basis
              To:   Finitely presented left module on 1 generator and 1 relation over mod 2 Steenrod algebra, milnor basis
              basis==True:
              []
            Hom(Trivial with redundant generator, Free):
              basis==False:
              Module morphism:
              From: Finitely presented left module on 1 generator and 1 relation over mod 2 Steenrod algebra, milnor basis
              To:   Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> 0
              basis==True:
              []
            Hom(Trivial with redundant generator, Hko):
              basis==False:
              Module morphism:
              From: Finitely presented left module on 1 generator and 1 relation over mod 2 Steenrod algebra, milnor basis
              To:   Finitely presented left module on 1 generator and 2 relations over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> 0
              basis==True:
              []
            Hom(Trivial with redundant generator, Trivial):
              basis==False:
              Module morphism:
              From: Finitely presented left module on 1 generator and 1 relation over mod 2 Steenrod algebra, milnor basis
              To:   Free graded left module on 0 generators over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> 0
              basis==True:
              []
            Hom(Trivial with redundant generator, Trivial with redundant generator):
              basis==False:
              Module endomorphism of Finitely presented left module on 1 generator and 1 relation over mod 2 Steenrod algebra, milnor basis
              Defn: g[0] |--> 0
              basis==True:
              []
        """
        from .morphism import _create_relations_matrix

        M = self.domain()
        N = self.codomain()

        def _trivial_case():
            '''
            The return value if there are no non-trivial homomorphisms.
            '''
            if basis:
                # Since the free module of homomorphisms is trivial, the basis
                # is the empty set.
                return []
            else:
                # Since the free module of homomorphisms is trivial, it contains
                # only The trivial homomorphism
                return self.zero()

        # Deal with the trivial cases first.  Note that this covers the case
        # where the domain or codomain have no generators.
        if N.is_trivial() or M.is_trivial():
            return _trivial_case()

        # Then deal with the case where the domain has no relations.
        elif not M.has_relations():
            res = []
            num_generators = len(M.generators())
            for i, g in enumerate(M.generators()):
                # The i'th generator can go to any of these basis elements:
                base = N.basis_elements(g.degree() + n)
                for value in base:
                    values = [N.zero() if i != j else value for j in range(num_generators)]
                    res.append(self(values))
                    if not basis:
                        return res[0]

        else:
            # Note that this list is non-empty since we dealt with the trivial
            # case above.
            source_degs = [g.degree() + n for g in M.generators()]

            # Note that this list is non-empty since we dealt with the free
            # case above.
            target_degs = [r.degree() + n for r in M.relations()]

            block_matrix, R = _create_relations_matrix(
                N,
                [r.dense_coefficient_list() for r in M.relations()],
                source_degs, target_degs)

            ker = R.right_kernel()

            res = []
            for b in ker.basis():
                n = 0

                xs = []
                for j, X in enumerate(block_matrix[0]):
                    k = X.domain().dimension()
                    xs.append(N.element_from_coordinates(b[n:n + k], source_degs[j]))
                    n += k

                res.append(self(xs))
                if not basis:
                    return res[0]

        # If the code above found a non-trivial homomorphism and ``basis==False``,
        # it will have terminated by now.
        if not res:
            return _trivial_case()
        else:
            return res



# sage_setup: distribution = sagemath-modules
"""
Utility Functions for Matrices

The actual computation of homology groups ends up being linear algebra
with the differentials thought of as matrices. This module contains
some utility functions for this purpose.
"""
########################################################################
#       Copyright (C) 2013 John H. Palmieri <palmieri@math.washington.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#
#                  https://www.gnu.org/licenses/
########################################################################

# TODO: this module is a clear candidate for cythonizing. Need to
# evaluate speed benefits.

from sage.matrix.constructor import matrix


def dhsw_snf(mat, verbose=False):
    """
    Preprocess a matrix using the "Elimination algorithm" described by
    Dumas et al. [DHSW2003]_, and then call ``elementary_divisors`` on the
    resulting (smaller) matrix.

    .. NOTE::

        'snf' stands for 'Smith Normal Form'.

    INPUT:

    - ``mat`` -- integer matrix, either sparse or dense

    (They use the transpose of the matrix considered here, so they use
    rows instead of columns.)

    ALGORITHM:

    Go through ``mat`` one column at a time.  For each
    column, add multiples of previous columns to it until either

    - it's zero, in which case it should be deleted.
    - its first nonzero entry is 1 or -1, in which case it should be kept.
    - its first nonzero entry is something else, in which case it is
      deferred until the second pass.

    Then do a second pass on the deferred columns.

    At this point, the columns with 1 or -1 in the first entry
    contribute to the rank of the matrix, and these can be counted and
    then deleted (after using the 1 or -1 entry to clear out its row).
    Suppose that there were `N` of these.

    The resulting matrix should be much smaller; we then feed it
    to Sage's ``elementary_divisors`` function, and prepend `N` 1s to
    account for the rows deleted in the previous step.

    EXAMPLES::

        sage: from sage.homology.matrix_utils import dhsw_snf
        sage: mat = matrix(ZZ, 3, 4, range(12))
        sage: dhsw_snf(mat)                                                             # needs sage.libs.pari
        [1, 4, 0]
        sage: mat = random_matrix(ZZ, 20, 20, x=-1, y=2)
        sage: mat.elementary_divisors() == dhsw_snf(mat)                                # needs sage.libs.pari
        True
    """
    ring = mat.base_ring()
    rows = mat.nrows()
    cols = mat.ncols()
    new_data = {}
    new_mat = matrix(ring, rows, cols, new_data)
    add_to_rank = 0
    zero_cols = 0
    if verbose:
        print("old matrix: %s by %s" % (rows, cols))
    # leading_positions: dictionary of lists indexed by row: if first
    # nonzero entry in column c is in row r, then leading_positions[r]
    # should contain c
    leading_positions = {}
    # pass 1:
    if verbose:
        print("starting pass 1")
    for j in range(cols):
        # new_col is a matrix with one column: sparse matrices seem to
        # be less buggy than sparse vectors (#5184, #5185), and
        # perhaps also faster.
        new_col = mat.matrix_from_columns([j])
        if new_col.is_zero():
            zero_cols += 1
        else:
            check_leading = True
            while check_leading:
                i = new_col.nonzero_positions_in_column(0)[0]
                entry = new_col[i, 0]
                check_leading = False
                if i in leading_positions:
                    for c in leading_positions[i]:
                        earlier = new_mat[i, c]
                        # right now we don't check to see if entry divides
                        # earlier, because we don't want to modify the
                        # earlier columns of the matrix.  Deal with this
                        # in pass 2.
                        if entry and earlier.divides(entry):
                            quo = entry.divide_knowing_divisible_by(earlier)
                            new_col = new_col - quo * new_mat.matrix_from_columns([c])
                            entry = 0
                            if not new_col.is_zero():
                                check_leading = True
            if not new_col.is_zero():
                new_mat.set_column(j-zero_cols, new_col.column(0))
                i = new_col.nonzero_positions_in_column(0)[0]
                if i in leading_positions:
                    leading_positions[i].append(j-zero_cols)
                else:
                    leading_positions[i] = [j-zero_cols]
            else:
                zero_cols += 1
    # pass 2:
    # first eliminate the zero columns at the end
    cols = cols - zero_cols
    zero_cols = 0
    new_mat = new_mat.matrix_from_columns(range(cols))
    if verbose:
        print("starting pass 2")
    keep_columns = list(range(cols))
    check_leading = True
    while check_leading:
        check_leading = False
        new_leading = leading_positions.copy()
        for i in leading_positions:
            if len(leading_positions[i]) > 1:
                j = leading_positions[i][0]
                jth = new_mat[i, j]
                for n in leading_positions[i][1:]:
                    nth = new_mat[i, n]
                    if jth.divides(nth):
                        quo = nth.divide_knowing_divisible_by(jth)
                        new_mat.add_multiple_of_column(n, j, -quo)
                    elif nth.divides(jth):
                        quo = jth.divide_knowing_divisible_by(nth)
                        jth = nth
                        new_mat.swap_columns(n, j)
                        new_mat.add_multiple_of_column(n, j, -quo)
                    else:
                        g, r, s = jth.xgcd(nth)
                        unit, A, B = r.xgcd(-s)  # unit ought to be 1 here
                        jth_col = new_mat.column(j)
                        nth_col = new_mat.column(n)
                        new_mat.set_column(j, r*jth_col + s*nth_col)
                        new_mat.set_column(n, B*jth_col + A*nth_col)
                        nth = B*jth + A*nth
                        jth = g
                        # at this point, jth should divide nth
                        quo = nth.divide_knowing_divisible_by(jth)
                        new_mat.add_multiple_of_column(n, j, -quo)
                    new_leading[i].remove(n)
                    if new_mat.column(n).is_zero():
                        keep_columns.remove(n)
                        zero_cols += 1
                    else:
                        new_r = new_mat.column(n).nonzero_positions()[0]
                        if new_r in new_leading:
                            new_leading[new_r].append(n)
                        else:
                            new_leading[new_r] = [n]
                        check_leading = True
        leading_positions = new_leading
    # pass 3: get rid of columns which start with 1 or -1
    if verbose:
        print("starting pass 3")
    max_leading = 1
    for i in leading_positions:
        j = leading_positions[i][0]
        entry = new_mat[i, j]
        if entry.abs() == 1:
            add_to_rank += 1
            keep_columns.remove(j)
            for c in new_mat.nonzero_positions_in_row(i):
                if c in keep_columns:
                    new_mat.add_multiple_of_column(c, j, -entry * new_mat[i, c])
        else:
            max_leading = max(max_leading, new_mat[i, j].abs())
    # form the new matrix
    if max_leading != 1:
        new_mat = new_mat.matrix_from_columns(keep_columns)
        if verbose:
            print("new matrix: %s by %s" % (new_mat.nrows(), new_mat.ncols()))
        if new_mat.is_sparse():
            ed = [1]*add_to_rank + new_mat.dense_matrix().elementary_divisors()
        else:
            ed = [1]*add_to_rank + new_mat.elementary_divisors()
    else:
        if verbose:
            print("new matrix: all pivots are 1 or -1")
        ed = [1]*add_to_rank

    if len(ed) < rows:
        return ed + [0]*(rows - len(ed))
    return ed[:rows]



# sage_setup: distribution = sagemath-modules
from sage.rings.finite_rings.stdint cimport *

cdef struct c_vector_modint:
    int_fast64_t *entries
    int p
    Py_ssize_t *positions
    Py_ssize_t degree
    Py_ssize_t num_nonzero

cdef int allocate_c_vector_modint(c_vector_modint* v, Py_ssize_t num_nonzero) except -1
cdef int init_c_vector_modint(c_vector_modint* v, int p, Py_ssize_t degree, Py_ssize_t num_nonzero) except -1
cdef void clear_c_vector_modint(c_vector_modint* v) noexcept
cdef Py_ssize_t binary_search0_modn(Py_ssize_t* v, Py_ssize_t n, int_fast64_t x) noexcept
cdef Py_ssize_t binary_search_modn(Py_ssize_t* v, Py_ssize_t n, int_fast64_t x, Py_ssize_t* ins) noexcept
cdef int_fast64_t get_entry(c_vector_modint* v, Py_ssize_t n) except -1
cdef bint is_entry_zero_unsafe(c_vector_modint* v, Py_ssize_t n) noexcept
cdef object to_list(c_vector_modint* v)
cdef int set_entry(c_vector_modint* v, Py_ssize_t n, int_fast64_t x) except -1
cdef int add_c_vector_modint_init(c_vector_modint* sum, c_vector_modint* v, c_vector_modint* w, int multiple) except -1
cdef int scale_c_vector_modint(c_vector_modint* v, int_fast64_t scalar) except -1



# sage_setup: distribution = sagemath-flint
from sage.libs.gmp.types cimport *
from sage.libs.flint.types cimport fmpz_mat_t

from sage.matrix.matrix_dense cimport Matrix_dense
from sage.rings.integer cimport Integer
from sage.ext.mod_int cimport *

ctypedef long* GEN

cdef class Matrix_integer_dense(Matrix_dense):
    cdef fmpz_mat_t _matrix
    cdef object _pivots
    cdef int mpz_height(self, mpz_t height) except -1
    cdef _mod_int_c(self, mod_int modulus)
    cdef _mod_two(self)
    cdef _pickle_version0(self)
    cdef _unpickle_version0(self, data)
    cpdef _export_as_string(self, int base=?)
    cdef void set_unsafe_mpz(self, Py_ssize_t i, Py_ssize_t j, const mpz_t value) noexcept
    cdef void set_unsafe_double(self, Py_ssize_t i, Py_ssize_t j, double value) noexcept
    cdef inline void get_unsafe_mpz(self, Py_ssize_t i, Py_ssize_t j, mpz_t value) noexcept
    cdef inline int get_unsafe_int(self, Py_ssize_t i, Py_ssize_t j) noexcept
    cdef inline double get_unsafe_double(self, Py_ssize_t i, Py_ssize_t j) noexcept

    # HNF Modn
    cdef int _hnf_modn(Matrix_integer_dense self, Matrix_integer_dense res,
            unsigned int det) except -1
    cdef int* _hnf_modn_impl(Matrix_integer_dense self, unsigned int det,
            Py_ssize_t nrows, Py_ssize_t ncols) except NULL

    cdef Matrix_integer_dense _new(self, Py_ssize_t nrows, Py_ssize_t ncols)



# sage_setup: distribution = sagemath-modules
r"""
Ternary quadratic form with integer coefficients

AUTHOR:

- Gustavo Rama

Based in code of Gonzalo Tornaria

The form `a\cdot x^2 + b\cdot y^2 + c\cdot z^2 + r\cdot yz + s\cdot xz + t\cdot xy` is stored as a tuple ``(a, b, c, r, s, t)`` of integers.
"""

# ****************************************************************************
#       Copyright (C) 2012 Gustavo Rama
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.misc import gcd, kronecker as kronecker_symbol
from sage.categories.rings import Rings
from sage.matrix.constructor import matrix, identity_matrix
from sage.misc.prandom import randint
from sage.quadratic_forms.quadratic_form import QuadraticForm
from sage.quadratic_forms.ternary import (_basic_lemma,
                                          _find_a_ternary_qf_by_level_disc,
                                          _find_all_ternary_qf_by_level_disc,
                                          _find_p_neighbor_from_vec,
                                          _find_zeros_mod_p_2,
                                          _find_zeros_mod_p_odd,
                                          _reduced_ternary_form_eisenstein_with_matrix,
                                          _reduced_ternary_form_eisenstein_without_matrix)

from sage.rings.finite_rings.integer_mod import mod
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.polynomial_ring import polygens
from sage.structure.element import Vector, Matrix
from sage.structure.sage_object import SageObject


class TernaryQF(SageObject):
    r"""
    The ``TernaryQF`` class represents a quadratic form in 3 variables with coefficients in `\ZZ`.

    INPUT:

    - ``v`` -- list or tuple of 6 entries:  ``[a,b,c,r,s,t]``

    OUTPUT: the ternary quadratic form `a\cdot x^2 + b\cdot y^2 + c\cdot z^2 + r\cdot y\cdot z + s\cdot x\cdot z + t\cdot x\cdot y`

    EXAMPLES::

        sage: Q = TernaryQF([1, 2, 3, 4, 5, 6]); Q
        Ternary quadratic form with integer coefficients:
        [1 2 3]
        [4 5 6]
        sage: A = matrix(ZZ, 3, [1, -7, 1, 0, -2, 1, 0, -1, 0])
        sage: Q(A)
        Ternary quadratic form with integer coefficients:
        [1 187 9]
        [-85 8 -31]
        sage: TestSuite(TernaryQF).run()
    """

    __slots__ = ['_a', '_b', '_c', '_r', '_s', '_t',
                 '_automorphisms', '_number_of_automorphisms']

    possible_automorphisms = None

    def __init__(self, v):
        r"""
        Create the ternary quadratic form `a\cdot x^2 + b\cdot y^2 + c\cdot z^2 + r\cdot y\cdot z + s\cdot x\cdot z + t\cdot x\cdot y` from the
        tuple ``v=[a,b,c,r,s,t]`` over `\ZZ`.

        INPUT:

        - ``v`` -- 6-tuple of integers

        EXAMPLES::

            sage: Q = TernaryQF([1, 2, 3, 4, 5, 6]); Q
            Ternary quadratic form with integer coefficients:
            [1 2 3]
            [4 5 6]
        """
        if len(v) != 6:
            # Check we have six coefficients
            raise ValueError("Ternary quadratic form must be given by a list of six coefficients")
        self._a, self._b, self._c, self._r, self._s, self._t = (ZZ(x) for x in v)
        self._automorphisms = None
        self._number_of_automorphisms = None

    def coefficients(self) -> tuple:
        r"""
        Return the tuple of coefficients of the ternary quadratic form.

        EXAMPLES::

            sage: Q = TernaryQF([1, 2, 3, 4, 5, 6]); Q
            Ternary quadratic form with integer coefficients:
            [1 2 3]
            [4 5 6]
            sage: Q.coefficients()
            (1, 2, 3, 4, 5, 6)
        """
        return self._a, self._b, self._c, self._r, self._s, self._t

    def __hash__(self) -> int:
        """
        Return a hash for ``self``.

        EXAMPLES::

            sage: Q = TernaryQF([1, 2, 3, 4, 5, 6])
            sage: Q.__hash__()
            5881802312257552497  # 64-bit
            1770036893           # 32-bit
        """
        return hash(self.coefficients())

    def coefficient(self, n):
        r"""
        Return the `n`-th coefficient of the ternary quadratic form.

        INPUT:

        - ``n`` -- integer with `0 \leq n \leq 5`

        EXAMPLES::

            sage: Q = TernaryQF([1, 2, 3, 4, 5, 6]); Q
            Ternary quadratic form with integer coefficients:
            [1 2 3]
            [4 5 6]
            sage: Q.coefficient(2)
            3
            sage: Q.coefficient(5)
            6
        """
        return self.coefficients()[n]

    def polynomial(self, names='x,y,z'):
        r"""
        Return the polynomial associated to the ternary quadratic form.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 0, 2, -3, -1]); Q
            Ternary quadratic form with integer coefficients:
            [1 1 0]
            [2 -3 -1]
            sage: p = Q.polynomial(); p
            x^2 - x*y + y^2 - 3*x*z + 2*y*z
            sage: p.parent()
            Multivariate Polynomial Ring in x, y, z over Integer Ring
        """
        x, y, z = polygens(ZZ, names)
        return self._a * x**2 + self._b * y**2 + self._c * z**2 + self._t * x*y + self._s * x*z + self._r * y*z

    def _repr_(self) -> str:
        r"""
        Display the quadratic form.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 0, 2, -3, -1])
            sage: print(Q._repr_())
            Ternary quadratic form with integer coefficients:
            [1 1 0]
            [2 -3 -1]
            sage: Q = TernaryQF([0, 0, 0, 0, 0, 0]); Q
            Ternary quadratic form with integer coefficients:
            [0 0 0]
            [0 0 0]
        """
        rep = 'Ternary quadratic form with integer coefficients:\n'
        rep += '[' + str(self._a) + ' ' + str(self._b) + ' ' + str(self._c) + ']\n'
        rep += '[' + str(self._r) + ' ' + str(self._s) + ' ' + str(self._t) + ']'
        return rep

    def __call__(self, v):
        r"""
        Evaluate this ternary quadratic form `Q` on a vector of 3 elements,
        or matrix of elements in Z, with 3 rows.

        OUTPUT:

        If a vector is given, then the output will be an integer `Q(v)`,
        but if a matrix is given, the output will be a ternary quadratic form
        if the matrix has 3 columns, or a quadratic form if not.
        The quadratic form in matrix notation will be:

        .. MATH::

            Q' = v^t\cdot Q\cdot v.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 1, -1, -2, -3])
            sage: Q((1, 1, 1))
            -3
            sage: M = matrix(ZZ, 3, 2, [358, 6, 2, 0, 0, 4])
            sage: Q(M)
            Quadratic form in 2 variables over Integer Ring with coefficients:
            [ 126020 1388 ]
            [ * 4 ]
            sage: M = matrix(ZZ, 3, 3, [1, 3, 0, -1, 4, 2, 1, -1, -1])
            sage: M
            [ 1  3  0]
            [-1  4  2]
            [ 1 -1 -1]
            sage: Q(M)
            Ternary quadratic form with integer coefficients:
            [5 0 7]
            [12 -13 -16]
        """
        if isinstance(v, Matrix):
            # Check that v has 3 rows
            if v.nrows() != 3:
                raise TypeError("the matrix must have 3 rows")
            # Check if v has 3 cols
            if v.ncols() == 3:
                M = v.transpose() * self.matrix() * v
                return TernaryQF([M[0, 0] // 2, M[1, 1] // 2, M[2, 2] // 2,
                                  M[1, 2], M[0, 2], M[0, 1]])

            return QuadraticForm(ZZ, v.transpose() * self.matrix() * v)
        if isinstance(v, (Vector, list, tuple)):
            # Check that v has length 3
            if len(v) != 3:
                raise TypeError("your vector needs to have length 3")
            v0, v1, v2 = v
            a, b, c, r, s, t = self.coefficients()
            return a*v0**2 + b*v1**2 + c*v2**2 + r*v1*v2 + s*v0*v2 + t*v0*v1

        raise TypeError("presently we can only evaluate a quadratic form on a list, tuple, vector or matrix")

    def quadratic_form(self):
        r"""
        Return a :class:`QuadraticForm` with the same coefficients as ``self`` over `\ZZ`.

        EXAMPLES::

            sage: Q = TernaryQF([1, 2, 3, 1, 1, 1])
            sage: QF1 = Q.quadratic_form(); QF1
            Quadratic form in 3 variables over Integer Ring with coefficients:
            [ 1 1 1 ]
            [ * 2 1 ]
            [ * * 3 ]
            sage: QF2 = QuadraticForm(ZZ, 3, [1, 1, 1, 2, 1, 3])
            sage: bool(QF1 == QF2)
            True
        """
        return QuadraticForm(ZZ, 3, [self._a, self._t, self._s,
                                     self._b, self._r, self._c])

    def matrix(self):
        r"""
        Return the Hessian matrix associated to the ternary quadratic form.

        That is, if `Q` is a ternary quadratic form, `Q(x,y,z) = a\cdot x^2 + b\cdot y^2 + c\cdot z^2 + r\cdot y\cdot z + s\cdot x\cdot z + t\cdot x\cdot y`,
        then the Hessian matrix associated to `Q` is
        ::

            [2\cdot a t s]
            [t 2\cdot b r]
            [s r 2\cdot c]

        EXAMPLES::

            sage: Q = TernaryQF([1,1,2,0,-1,4]); Q
            Ternary quadratic form with integer coefficients:
            [1 1 2]
            [0 -1 4]
            sage: M = Q.matrix(); M
            [ 2  4 -1]
            [ 4  2  0]
            [-1  0  4]
            sage: v = vector((1, 2, 3))
            sage: Q(v)
            28
            sage: (v*M*v.column())[0]//2
            28
        """
        return matrix(ZZ, 3, 3, [2 * self._a, self._t, self._s,
                                 self._t, 2 * self._b, self._r,
                                 self._s, self._r, 2 * self._c])

    def disc(self):
        r"""
        Return the discriminant of the ternary quadratic form.

        This is the determinant of the matrix divided by 2.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 2, 0, -1, 4])
            sage: Q.disc()
            -25
            sage: Q.matrix().det()
            -50
        """
        return (4*self._a*self._b*self._c + self._r*self._s*self._t
                - self._a*self._r**2 - self._b*self._s**2 - self._c*self._t**2)

    def is_definite(self) -> bool:
        """
        Determine if the ternary quadratic form is definite.

        EXAMPLES::

            sage: Q = TernaryQF([10, 10, 1, -1, 2, 3])
            sage: Q.is_definite()
            True
            sage: (-Q).is_definite()
            True
            sage: Q = TernaryQF([1, 1, 2, -3, 0, -1])
            sage: Q.is_definite()
            False
        """
        d1 = self._a
        if d1 == 0:
            return False
        d2 = 4 * self._a * self._b - self._t**2
        if d2 <= 0:
            return False
        d3 = self.disc()
        if d3 == 0:
            return False
        return (d1 > 0) == (d3 > 0)

    def is_positive_definite(self) -> bool:
        """
        Determine if the ternary quadratic form is positive definite.

        EXAMPLES::

            sage: Q = TernaryQF([10, 10, 1, -1, 2, 3])
            sage: Q.is_positive_definite()
            True
            sage: (-Q).is_positive_definite()
            False
            sage: Q = TernaryQF([1, 1, 0, 0, 0, 0])
            sage: Q.is_positive_definite()
            False
            sage: Q = TernaryQF([1, 1, 1, -1, -2, -3])
            sage: Q((1,1,1))
            -3
            sage: Q.is_positive_definite()
            False
        """
        d1 = self._a
        if d1 == 0:
            return False
        d2 = 4 * self._a * self._b - self._t**2
        if d2 <= 0:
            return False
        d3 = self.disc()
        if d3 == 0:
            return False
        return d1 > 0 and d3 > 0

    def is_negative_definite(self) -> bool:
        """
        Determine if the ternary quadratic form is negative definite.

        EXAMPLES::

            sage: Q = TernaryQF([-8, -9, -10, 1, 9, -3])
            sage: Q.is_negative_definite()
            True
            sage: Q = TernaryQF([-4, -1, 6, -5, 1, -5])
            sage: Q((0, 0, 1))
            6
            sage: Q.is_negative_definite()
            False
        """
        d1 = self._a
        if d1 == 0:
            return False
        d2 = 4 * self._a * self._b - self._t**2
        if d2 <= 0:
            return False
        d3 = self.disc()
        if d3 == 0:
            return False
        return d1 < 0 and d3 < 0

    def __neg__(self):
        """
        Return the ternary quadratic form with coefficients negatives of ``self``.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 2, -2, 0, -1]); Q
            Ternary quadratic form with integer coefficients:
            [1 1 2]
            [-2 0 -1]
            sage: -Q
            Ternary quadratic form with integer coefficients:
            [-1 -1 -2]
            [2 0 1]
            sage: Q = TernaryQF([0, 0, 0, 0, 0, 0])
            sage: Q == -Q
            True
        """
        return TernaryQF([-a for a in self.coefficients()])

    def is_primitive(self) -> bool:
        """
        Determine if the ternary quadratic form is primitive.

        This means that the greatest common divisor of the coefficients
        of the form is 1.

        EXAMPLES::

            sage: Q = TernaryQF([1, 2, 3, 4, 5, 6])
            sage: Q.is_primitive()
            True
            sage: Q.content()
            1
            sage: Q = TernaryQF([10, 10, 10, 5, 5, 5])
            sage: Q.content()
            5
            sage: Q.is_primitive()
            False
        """
        return self.content() == 1

    def primitive(self):
        """
        Return the primitive version of the ternary quadratic form.

        EXAMPLES::

            sage: Q = TernaryQF([2, 2, 2, 1, 1, 1])
            sage: Q.is_primitive()
            True
            sage: Q.primitive()
            Ternary quadratic form with integer coefficients:
            [2 2 2]
            [1 1 1]
            sage: Q.primitive() == Q
            True
            sage: Q = TernaryQF([10, 10, 10, 5, 5, 5])
            sage: Q.primitive()
            Ternary quadratic form with integer coefficients:
            [2 2 2]
            [1 1 1]
        """
        l = self.coefficients()
        g = gcd(l)
        return TernaryQF([a // g for a in l])

    def scale_by_factor(self, k):
        """
        Scale the values of the ternary quadratic form by the number ``k``.

        OUTPUT:

        If ``k`` times the content of the ternary quadratic form is an integer, return a ternary quadratic form;
        otherwise, return a quadratic form of dimension 3.

        EXAMPLES::

            sage: Q = TernaryQF([2, 2, 4, 0, -2, 8])
            sage: Q
            Ternary quadratic form with integer coefficients:
            [2 2 4]
            [0 -2 8]
            sage: Q.scale_by_factor(5)
            Ternary quadratic form with integer coefficients:
            [10 10 20]
            [0 -10 40]
            sage: Q.scale_by_factor(1/2)
            Ternary quadratic form with integer coefficients:
            [1 1 2]
            [0 -1 4]
            sage: Q.scale_by_factor(1/3)
            Quadratic form in 3 variables over Rational Field with coefficients:
            [ 2/3 8/3 -2/3 ]
            [ * 2/3 0 ]
            [ * * 4/3 ]
        """
        if k * self.content() in ZZ:
            return TernaryQF([ZZ(k*self._a), ZZ(k*self._b), ZZ(k*self._c),
                              ZZ(k*self._r), ZZ(k*self._s), ZZ(k*self._t)])

        R = k.parent()
        if R not in Rings():
            raise TypeError(f"{k} does not belong to a ring")

        return QuadraticForm(R, 3, [k * self._a, k * self._t, k * self._s,
                                    k * self._b, k * self._r, k * self._c])

    def reciprocal(self):
        """
        Return the reciprocal quadratic form associated to the given form.

        This is defined as the multiple of the primitive adjoint with the same
        content as the given form.

        EXAMPLES::

            sage: Q = TernaryQF([2, 2, 14, 0, 0, 0])
            sage: Q.reciprocal()
            Ternary quadratic form with integer coefficients:
            [14 14 2]
            [0 0 0]
            sage: Q.content()
            2
            sage: Q.reciprocal().content()
            2
            sage: Q.adjoint().content()
            16
        """
        return self.adjoint().primitive().scale_by_factor(self.content())

    def reciprocal_reduced(self):
        """
        Return the reduced form of the reciprocal form of the given ternary quadratic form.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 3, 0, -1, 0])
            sage: Qrr = Q.reciprocal_reduced(); Qrr
            Ternary quadratic form with integer coefficients:
            [4 11 12]
            [0 -4 0]
            sage: Q.is_eisenstein_reduced()
            True
            sage: Qr = Q.reciprocal()
            sage: Qr.reduced_form_eisenstein(matrix=False) == Qrr
            True
        """
        return self.reciprocal().reduced_form_eisenstein(matrix=False)

    def divisor(self):
        """
        Return the content of the adjoint form associated to the given form.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 17, 0, 0, 0])
            sage: Q.divisor()
            4
        """
        A11 = 4*self._b*self._c - self._r**2
        A22 = 4*self._a*self._c - self._s**2
        A33 = 4*self._a*self._b - self._t**2
        A23 = self._s*self._t - 2*self._a*self._r
        A13 = self._r*self._t - 2*self._b*self._s
        A12 = self._r*self._s - 2*self._c*self._t
        m = gcd([A11, A22, A33, 2*A12, 2*A13, 2*A23])
        return m

    def __eq__(self, right) -> bool:
        """
        Determine if two ternary quadratic forms are equal.

        EXAMPLES::

            sage: Q = TernaryQF([1, 2, 3, 1, 2, 3])
            sage: Q == Q
            True
            sage: Q1 = TernaryQF([1, 2, 3, 1, 2, 2])
            sage: Q == Q1
            False
        """
        if not isinstance(right, TernaryQF):
            return False
        return self.coefficients() == right.coefficients()

    def adjoint(self):
        """
        Return the adjoint form associated to the given ternary quadratic
        form.

        That is, the Hessian matrix of the adjoint form is twice the
        classical adjoint matrix of the Hessian matrix of the given form.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 17, 0, 0, 1])
            sage: Q.adjoint()
            Ternary quadratic form with integer coefficients:
            [68 68 3]
            [0 0 -68]
            sage: Q.adjoint().matrix() == 2*Q.matrix().adjoint_classical()              # needs sage.libs.pari
            True
        """
        A11 = 4*self._b*self._c - self._r**2
        A22 = 4*self._a*self._c - self._s**2
        A33 = 4*self._a*self._b - self._t**2
        A23 = self._s*self._t - 2*self._a*self._r
        A13 = self._r*self._t - 2*self._b*self._s
        A12 = self._r*self._s - 2*self._c*self._t
        return TernaryQF([A11, A22, A33, 2*A23, 2*A13, 2*A12])

    def content(self):
        """
        Return the greatest common divisor of the coefficients of the given ternary quadratic form.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 2, 0, 0, 0])
            sage: Q.content()
            1
            sage: Q = TernaryQF([2, 4, 6, 0, 0, 0])
            sage: Q.content()
            2
            sage: Q.scale_by_factor(100).content()
            200
        """
        return gcd(self.coefficients())

    def omega(self):
        """
        Return the content of the adjoint of the primitive associated
        ternary quadratic form.

        EXAMPLES::

            sage: Q = TernaryQF([4, 11, 12, 0, -4, 0])
            sage: Q.omega()
            176
            sage: Q.primitive().adjoint().content()
            176
        """
        return self.primitive().adjoint().content()

    def delta(self):
        """
        Return the omega of the adjoint of the given ternary quadratic form,
        which is the same as the omega of the reciprocal form.

        EXAMPLES::

            sage: Q = TernaryQF([1, 2, 2, -1, 0, -1])
            sage: Q.delta()
            208
            sage: Q.adjoint().omega()
            208
            sage: Q = TernaryQF([1, -1, 1, 0, 0, 0])
            sage: Q.delta()
            4
            sage: Q.omega()
            4
        """
        return self.adjoint().omega()

    def level(self):
        """
        Return the level of the ternary quadratic form, which is 4 times the discriminant divided by the divisor.

        EXAMPLES::

            sage: Q = TernaryQF([1, 2, 2, -1, 0, -1])
            sage: Q.level()
            52
            sage: 4*Q.disc()/Q.divisor()
            52
        """
        return 4 * self.disc() // self.divisor()

    def is_eisenstein_reduced(self) -> bool:
        r"""
        Determine if the ternary quadratic form is Eisenstein reduced.

        That is, if we have a ternary quadratic form:
        ::

        [a b c]
        [r s t]

        then

        1. `a \leq b \leq c`;
        2. `r`, `s`, and `t` are all positive or all nonpositive;
        3. `a \geq |t|`; `a \geq |s|`; `b \geq |r|`;
        4. `a+b+r+s+t \geq 0`;
        5. `a=t` implies `s \leq 2\cdot r`; `a=s` implies `t \leq 2\cdot r`; `b=r` implies `t \leq 2\cdot s`;
        6. `a=-t` implies `s=0`; `a=-s` implies `t=0`; `b=-r` implies `t=0`;
        7. `a+b+r+s+t = 0` implies `2\cdot a+2\cdot s+t \leq 0`;
        8. `a=b` implies `|r| \leq |s|`; `b=c` implies `|s| \leq |t|`.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 1, 0, 0, 0])
            sage: Q.is_eisenstein_reduced()
            True
            sage: Q = TernaryQF([34, 14, 44, 12, 25, -22])
            sage: Q.is_eisenstein_reduced()
            False
        """
        a, b, c, r, s, t = [self._a, self._b, self._c,
                            self._r, self._s, self._t]

        # cond 2
        if not (r > 0 and t > 0 and s > 0):
            if not (r <= 0 and s <= 0 and t <= 0):
                return False

        # cond 1 & 4
        if not (a <= b <= c and 0 <= a+b+r+s+t):
            return False

        # cond 3
        if not (a >= abs(s) and a >= abs(t) and b >= abs(r)):
            return False

        # cond 8
        if a == b and abs(r) > abs(s):
            return False
        if b == c and abs(s) > abs(t):
            return False
        if a+b+r+s+t == 0 and 2*a+2*s+t > 0:
            return False

        # cond 6
        # r, s, t <= 0
        if r <= 0:
            if a == -t and s != 0:
                return False
            if a == -s and t != 0:
                return False
            if b == -r and t != 0:
                return False

        # cond 7
        # r, s, t > 0
        if a == t and s > 2*r:
            return False
        if a == s and t > 2*r:
            return False
        return not (b == r and t > 2 * s)

    def reduced_form_eisenstein(self, matrix=True):
        r"""
        Return the Eisenstein reduced form equivalent to the given positive ternary quadratic form,
        which is unique.

        EXAMPLES::

            sage: Q = TernaryQF([293, 315, 756, 908, 929, 522])
            sage: Qr, m = Q.reduced_form_eisenstein()
            sage: Qr
            Ternary quadratic form with integer coefficients:
            [1 2 2]
            [-1 0 -1]
            sage: Qr.is_eisenstein_reduced()
            True
            sage: m
            [ -54  137  -38]
            [ -23   58  -16]
            [  47 -119   33]
            sage: m.det()
            1
            sage: Q(m) == Qr
            True
            sage: Q = TernaryQF([12,36,3,14,-7,-19])
            sage: Q.reduced_form_eisenstein(matrix = False)
            Ternary quadratic form with integer coefficients:
            [3 8 20]
            [3 2 1]
        """
        if matrix:
            v, M = _reduced_ternary_form_eisenstein_with_matrix(self._a, self._b, self._c, self._r, self._s, self._t)
            return TernaryQF(v), M

        v = _reduced_ternary_form_eisenstein_without_matrix(self._a, self._b, self._c, self._r, self._s, self._t)
        return TernaryQF(v)

    def pseudorandom_primitive_zero_mod_p(self, p):
        """
        Return a tuple of the form `v = (a, b, 1)` such that is a zero of the given ternary quadratic
        positive definite form modulo an odd prime `p`, where `p` doesn't divides the discriminant of the form.

        EXAMPLES::

             sage: Q = TernaryQF([1, 1, 11, 0, -1, 0])
             sage: Q.disc()
             43
             sage: Q.pseudorandom_primitive_zero_mod_p(3)  # random
             (1, 2, 1)
             sage: Q((1, 2, 1))
             15
             sage: v = Q.pseudorandom_primitive_zero_mod_p(1009)                        # needs sage.libs.pari
             sage: Q(v) % 1009                                                          # needs sage.libs.pari
             0
             sage: v[2]                                                                 # needs sage.libs.pari
             1
        """
        a, b, c, r, s, t = self.coefficients()
        while True:

            r1 = randint(0, p-1)
            r2 = randint(0, p-1)
            alpha = (b*r1**2+t*r1+a) % p
            if alpha != 0:

                beta = (2*b*r1*r2+t*r2+r*r1+s) % p
                gamma = (b*r2**2+r*r2+c) % p
                disc = beta**2-4*alpha*gamma
                if mod(disc, p).is_square():

                    z = (-beta+mod(disc, p).sqrt().lift())*(2*alpha).inverse_mod(p)
                    # return vector((z,r1*z+r2,1))%p
                    return z % p, (r1*z+r2) % p, 1

    def find_zeros_mod_p(self, p):
        """
        Find the zeros of the given ternary quadratic positive definite form modulo a prime `p`, where `p` doesn't divide the discriminant of the form.

        EXAMPLES::

            sage: Q = TernaryQF([4, 7, 8, -4, -1, -3])
            sage: Q.is_positive_definite()
            True
            sage: Q.disc().factor()
            3 * 13 * 19
            sage: Q.find_zeros_mod_p(2)
            [(1, 0, 0), (1, 1, 0), (0, 0, 1)]
            sage: zeros_17 = Q.find_zeros_mod_p(17)                                     # needs sage.libs.pari
            sage: len(zeros_17)                                                         # needs sage.libs.pari
            18
            sage: [Q(v)%17 for v in zeros_17]                                           # needs sage.libs.pari
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        """
        if p == 2:
            return _find_zeros_mod_p_2(self._a, self._b, self._c,
                                       self._r, self._s, self._t)

        v = self.pseudorandom_primitive_zero_mod_p(p)
        a, b, c, r, s, t = self.coefficients()
        return _find_zeros_mod_p_odd(a, b, c, r, s, t, p, v)

    def find_p_neighbor_from_vec(self, p, v, mat=False):
        r"""
        Finds the reduced equivalent of the `p`-neighbor of this ternary quadratic form associated to a given
        vector `v` satisfying:

        1. `Q(v) = 0`  (mod `p`)

        2. `v` is a non-singular point of the conic `Q(v) = 0` (mod `p`).

        REFERENCES:

        Gonzalo Tornaria's Thesis, Thrm 3.5, p34.

        EXAMPLES::

            sage: # needs sage.libs.pari
            sage: Q = TernaryQF([1, 3, 3, -2, 0, -1]); Q
            Ternary quadratic form with integer coefficients:
            [1 3 3]
            [-2 0 -1]
            sage: Q.disc()
            29
            sage: v = (9, 7, 1)
            sage: v in Q.find_zeros_mod_p(11)
            True
            sage: Q11, M = Q.find_p_neighbor_from_vec(11, v, mat=True)
            sage: Q11
            Ternary quadratic form with integer coefficients:
            [1 2 4]
            [-1 -1 0]
            sage: M
            [    -1  -5/11   7/11]
            [     0 -10/11   3/11]
            [     0  -3/11  13/11]
            sage: Q(M) == Q11
            True

        Test that it works with (0, 0, 1)::

            sage: Q.find_p_neighbor_from_vec(3, (0,0,1))                                # needs sage.libs.pari
            Ternary quadratic form with integer coefficients:
            [1 3 3]
            [-2 0 -1]
        """
        if mat:
            q, M = _find_p_neighbor_from_vec(self._a, self._b, self._c, self._r, self._s, self._t, p, v, mat)
            M = matrix(3, M)
            return TernaryQF(q), M*M.det()
        else:
            return TernaryQF(_find_p_neighbor_from_vec(self._a, self._b, self._c, self._r, self._s, self._t, p, v, mat))

    def find_p_neighbors(self, p, mat=False):
        """
        Find a list with all the reduced equivalent of the `p`-neighbors of this ternary quadratic form, given by the zeros mod `p` of the form.
        See :meth:`find_p_neighbor_from_vec` for more information.

        EXAMPLES::

            sage: # needs sage.libs.pari
            sage: Q0 = TernaryQF([1, 3, 3, -2, 0, -1]); Q0
            Ternary quadratic form with integer coefficients:
            [1 3 3]
            [-2 0 -1]
            sage: neig = Q0.find_p_neighbors(5)
            sage: len(neig)
            6
            sage: Q1 = TernaryQF([1, 1, 10, 1, 1, 1])
            sage: Q2 = TernaryQF([1, 2, 4, -1, -1, 0])
            sage: neig.count(Q0)
            2
            sage: neig.count(Q1)
            1
            sage: neig.count(Q2)
            3
        """
        z = self.find_zeros_mod_p(p)
        return [self.find_p_neighbor_from_vec(p, v, mat) for v in z]

    def basic_lemma(self, p):
        """
        Find a number represented by ``self`` and coprime to the prime `p`.

        EXAMPLES::

            sage: Q = TernaryQF([3, 3, 3, -2, 0, -1])
            sage: Q.basic_lemma(3)
            4
        """
        return _basic_lemma(self._a, self._b, self._c, self._r, self._s, self._t, p)

    def xi(self, p):
        """
        Return the value of the genus characters Xi_p... which may be
        missing one character. We allow `-1` as a prime.

        REFERENCES:

        Dickson's "Studies in the Theory of Numbers"

        EXAMPLES::

            sage: Q1 = TernaryQF([26, 42, 53, -36, -17, -3])
            sage: Q2 = Q1.find_p_neighbors(2)[1]
            sage: Q1.omega()
            3
            sage: Q1.xi(3), Q2.xi(3)
            (-1, -1)
        """
        if p == 4:
            p = -1
        if p == 8:
            p = 2

        if self.omega() % p != 0:
            raise ValueError("not a valid character")

        if p == -1 and self.omega() % 2**4 != 0:
            raise ValueError("not a valid character")

        if p == 2 and self.omega() % 2**5 != 0:
            raise ValueError("not a valid character")

        if (p == -1) or (p == 2):
            return kronecker_symbol(p, self.basic_lemma(2))

        return kronecker_symbol(self.basic_lemma(p), p)

    def xi_rec(self, p):
        """
        Return Xi(p) for the reciprocal form.

        EXAMPLES::

            sage: Q1 = TernaryQF([1, 1, 7, 0, 0, 0])
            sage: Q2 = Q1.find_p_neighbors(3)[0]
            sage: Q1.delta()
            28
            sage: Q1.xi_rec(7), Q2.xi_rec(7)
            (1, 1)
        """
        return self.reciprocal().xi(p)

    def symmetry(self, v):
        """
        Return `A`, the automorphism of the ternary quadratic form such that:

        - `Av = -v`,
        - `Au = 0`, if `u` is orthogonal to `v`,

        where `v` is a given vector.

        EXAMPLES::

            sage: Q = TernaryQF([4, 5, 8, 5, 2, 2])
            sage: v = vector((1,1,1))
            sage: M = Q.symmetry(v)
            sage: M
            [  7/13 -17/26 -23/26]
            [ -6/13   9/26 -23/26]
            [ -6/13 -17/26   3/26]
            sage: M.det()
            -1
            sage: M*v
            (-1, -1, -1)
            sage: v1 = vector((23, 0, -12))
            sage: v2 = vector((0, 23, -17))
            sage: v1*Q.matrix()*v
            0
            sage: v2*Q.matrix()*v
            0
            sage: M*v1 == v1
            True
            sage: M*v2 == v2
            True
        """

        return identity_matrix(3) - v.column()*matrix(v)*self.matrix()/self(v)

    def automorphism_symmetries(self, A) -> list:
        """
        Given the automorphism `A`, if `A` is the identity, return the empty list.
        Otherwise, return a list of two vectors `v_1`, `v_2` such that the product of
        the symmetries of the ternary quadratic form given by the two vectors is `A`.

        EXAMPLES::

            sage: Q = TernaryQF([9, 12, 30, -26, -28, 20])
            sage: A = matrix(ZZ, 3, [9, 10, -10, -6, -7, 6, 2, 2, -3])
            sage: Q(A) == Q
            True
            sage: v1, v2 = Q.automorphism_symmetries(A)
            sage: v1, v2
            ((8, -6, 2), (1, -5/4, -1/4))
            sage: A1 = Q.symmetry(v1)
            sage: A1
            [    9     9   -13]
            [   -6 -23/4  39/4]
            [    2   9/4  -9/4]
            sage: A2 = Q.symmetry(v2)
            sage: A2
            [    1     1     3]
            [    0  -1/4 -15/4]
            [    0  -1/4   1/4]
            sage: A1*A2 == A
            True
            sage: Q.automorphism_symmetries(identity_matrix(ZZ,3))
            []
        """
        if A == identity_matrix(3):
            return []

        bs = (A - 1).columns()
        b1 = next(v for v in bs if v)
        A1 = self.symmetry(b1) * A
        bs = (A1 - 1).columns()
        b2 = next(v for v in bs if v)
        return [b1, b2]

    def automorphism_spin_norm(self, A):
        """
        Return the spin norm of the automorphism `A`.

        EXAMPLES::

            sage: Q = TernaryQF([9, 12, 30, -26, -28, 20])
            sage: A = matrix(ZZ, 3, [9, 10, -10, -6, -7, 6, 2, 2, -3])
            sage: A.det()
            1
            sage: Q(A) == Q
            True
            sage: Q.automorphism_spin_norm(A)
            7
        """
        if A == identity_matrix(ZZ, 3):
            return 1
        b1, b2 = self.automorphism_symmetries(A)
        s = self(b1) * self(b2)
        return s.squarefree_part()

    def _border(self, n) -> bool:
        """
        Auxiliary function to find the automorphisms of a positive definite ternary quadratic form.

        It returns a boolean whether the n-condition is true.

        If ``Q = TernaryQF([a,b,c,r,s,t])``, the conditions are:

        1.  a = t, s = 2r.
        2.  a = s, t = 2r.
        3.  b = r, t = 2s.
        4.  a = -t.
        5.  a = -s.
        6.  b = -r.
        7.  a + b + r + s + t = 0, 2a + 2s + t = 0.
        8.  a = b, r = s.
        9.  b = c, s = t.
        10. r = s, r = 0.
        11. r = t, r = 0.
        12. s = t, s = 0.
        13. r = s, s = t, t = a.
        14. a = s, a = t.
        15. a = b, a + b + r + s + t = 0.
        16. a = b, b = c, a + b + r + s + t = 0.

        EXAMPLES::

            sage: Q01 = TernaryQF([5, 5, 9, 2, 4, 5])
            sage: Q01._border(1)
            True
            sage: Q02 = TernaryQF([6, 7, 8, 2, 6, 4])
            sage: Q02._border(2)
            True
            sage: Q03 = TernaryQF([6, 9, 9, 9, 3, 6])
            sage: Q03._border(3)
            True
            sage: Q04 = TernaryQF([1, 2, 3, -1, 0, -1])
            sage: Q04._border(4)
            True
            sage: Q05 = TernaryQF([2, 3, 5, -1, -2, 0])
            sage: Q05._border(5)
            True
            sage: Q06 = TernaryQF([1, 5, 7, -5, 0, 0])
            sage: Q06._border(6)
            True
            sage: Q07 = TernaryQF([1, 1, 7, -1, -1, 0])
            sage: Q07._border(7)
            True
            sage: Q08 = TernaryQF([2, 2, 5, -1, -1, -1])
            sage: Q08._border(8)
            True
            sage: Q09 = TernaryQF([3, 8, 8, 6, 2, 2])
            sage: Q09._border(9)
            True
            sage: Q10 = TernaryQF([1, 3, 4, 0, 0, 0])
            sage: Q10._border(10)
            True
            sage: Q11 = TernaryQF([3, 5, 8, 0, -1, 0])
            sage: Q11._border(11)
            True
            sage: Q12 = TernaryQF([2, 6, 7, -5, 0, 0])
            sage: Q12._border(12)
            True
            sage: Q13 = TernaryQF([1, 1, 2, 1, 1, 1])
            sage: Q13._border(13)
            True
            sage: Q14 = TernaryQF([1, 3, 4, 3, 1, 1])
            sage: Q14._border(14)
            True
            sage: Q15 = TernaryQF([3, 3, 6, -3, -3, 0])
            sage: Q15._border(15)
            True
            sage: Q16 = TernaryQF([4, 4, 4, -2, -3, -3])
            sage: Q16._border(16)
            True
        """
        a, b, c, r, s, t = self.coefficients()
        if n == 1:
            return (a == t) and (s == 2*r)
        elif n == 2:
            return (a == s) and (t == 2*r)
        elif n == 3:
            return (b == r) and (t == 2*s)
        elif n == 4:
            return (a == -t)
        elif n == 5:
            return (a == -s)
        elif n == 6:
            return (b == -r)
        elif n == 7:
            return (a + b + r + s + t == 0) and (2*a + 2*s + t == 0)
        elif n == 8:
            return (a == b) and (r == s)
        elif n == 9:
            return (b == c) and (s == t)
        elif n == 10:
            return (r == s) and (r == 0)
        elif n == 11:
            return (r == t) and (r == 0)
        elif n == 12:
            return (s == t) and (s == 0)
        elif n == 13:
            return (r == s) and (s == t) and (t == a)
        elif n == 14:
            return (a == s) and (a == t)
        elif n == 15:
            return (a == b) and (a + b + r + s + t == 0)
        elif n == 16:
            return (a == b) and (b == c) and (a + b + r + s + t == 0)

    def _borders(self):
        """
        Return the borders that the ternary quadratic form meet.

        See: TernaryQF._border

        EXAMPLES::

            sage: Q01 = TernaryQF([5, 5, 9, 2, 4, 5])
            sage: Q01._borders()
            (1,)
            sage: Q02 = TernaryQF([6, 7, 8, 2, 6, 4])
            sage: Q02._borders()
            (2,)
            sage: Q03 = TernaryQF([6, 9, 9, 9, 3, 6])
            sage: Q03._borders()
            (3,)
            sage: Q04 = TernaryQF([1, 2, 3, -1, 0, -1])
            sage: Q04._borders()
            (4,)
            sage: Q05 = TernaryQF([2, 3, 5, -1, -2, 0])
            sage: Q05._borders()
            (5,)
            sage: Q06 = TernaryQF([1, 5, 7, -5, 0, 0])
            sage: Q06._borders()
            (6, 12)
            sage: Q07 = TernaryQF([1, 1, 7, -1, -1, 0])
            sage: Q07._borders()
            (5, 6, 7, 8, 15)
            sage: Q08 = TernaryQF([2, 2, 5, -1, -1, -1])
            sage: Q08._borders()
            (8,)
            sage: Q09 = TernaryQF([3, 8, 8, 6, 2, 2])
            sage: Q09._borders()
            (9,)
            sage: Q10 = TernaryQF([1, 3, 4, 0, 0, 0])
            sage: Q10._borders()
            (10, 11, 12)
            sage: Q11 = TernaryQF([3, 5, 8, 0, -1, 0])
            sage: Q11._borders()
            (11,)
            sage: Q12 = TernaryQF([2, 6, 7, -5, 0, 0])
            sage: Q12._borders()
            (12,)
            sage: Q13 = TernaryQF([1, 1, 2, 1, 1, 1])
            sage: Q13._borders()
            (8, 13, 14)
            sage: Q14 = TernaryQF([1, 3, 4, 3, 1, 1])
            sage: Q14._borders()
            (14,)
            sage: Q15 = TernaryQF([3, 3, 6, -3, -3, 0])
            sage: Q15._borders()
            (5, 6, 7, 8, 15)
            sage: Q16 = TernaryQF([4, 4, 4, -2, -3, -3])
            sage: Q16._borders()
            (9, 15, 16)
        """
        return tuple(n for n in range(1, 17) if self._border(n))

    def _automorphisms_reduced_fast(self):
        """
        Return the coefficients of the matrices of the automorphisms of the reduced ternary quadratic form.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 7, 0, 0, 0])
            sage: Q.is_eisenstein_reduced()
            True
            sage: auts = Q._automorphisms_reduced_fast()
            sage: len(auts)
            8
            sage: A = matrix(3, auts[randint(0,7)])
            sage: Q(A) == Q
            True
            sage: Q = TernaryQF([3, 4, 5, 3, 3, 2])
            sage: Q._automorphisms_reduced_fast()
            [(1, 0, 0, 0, 1, 0, 0, 0, 1)]
        """
        if self._border(1):
            if self._border(2):
                if self._border(14):
                    if self._border(9):
                        # borders 1, 2, 9, 14
                        return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                                (-1, -1, -1, 0, 0, 1, 0, 1, 0),
                                (-1, -1, 0, 0, 1, 0, 0, 0, -1),
                                (-1, 0, -1, 0, -1, 0, 0, 0, 1),
                                (-1, 0, 0, 0, 0, -1, 0, -1, 0),
                                (1, 0, 1, 0, 0, -1, 0, 1, 0),
                                (1, 1, 0, 0, 0, 1, 0, -1, 0),
                                (1, 1, 1, 0, -1, 0, 0, 0, -1)]
                    else:
                        # borders 1, 2, 14
                        return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                                (-1, -1, 0, 0, 1, 0, 0, 0, -1),
                                (-1, 0, -1, 0, -1, 0, 0, 0, 1),
                                (1, 1, 1, 0, -1, 0, 0, 0, -1)]
            else:
                # borders 1
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (-1, -1, 0, 0, 1, 0, 0, 0, -1)]

        if self._border(2):
            # borders 2
            return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                    (-1, 0, -1, 0, -1, 0, 0, 0, 1)]

        if self._border(3):
            # borders 3
            return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                    (-1, 0, 0, 0, -1, -1, 0, 0, 1)]

        if self._border(4):
            if self._border(10):
                if self._border(8):
                    # borders 4, 8, 10
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, 0, 0, -1, 1, 0, 0, 0, -1),
                            (-1, 0, 0, 0, -1, 0, 0, 0, 1),
                            (-1, 1, 0, -1, 0, 0, 0, 0, 1),
                            (-1, 1, 0, 0, 1, 0, 0, 0, -1),
                            (0, -1, 0, -1, 0, 0, 0, 0, -1),
                            (0, -1, 0, 1, -1, 0, 0, 0, 1),
                            (0, 1, 0, -1, 1, 0, 0, 0, 1),
                            (0, 1, 0, 1, 0, 0, 0, 0, -1),
                            (1, -1, 0, 0, -1, 0, 0, 0, -1),
                            (1, -1, 0, 1, 0, 0, 0, 0, 1),
                            (1, 0, 0, 1, -1, 0, 0, 0, -1)]
                else:
                    # borders 4, 10
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, 0, 0, 0, -1, 0, 0, 0, 1),
                            (-1, 1, 0, 0, 1, 0, 0, 0, -1),
                            (1, -1, 0, 0, -1, 0, 0, 0, -1)]
            else:
                # borders 4
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (1, -1, 0, 0, -1, 0, 0, 0, -1)]

        if self._border(5):
            if self._border(6):
                if self._border(7):
                    if self._border(8):
                        if self._border(15):
                            # borders 5, 6, 7, 8, 15
                            return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                                    (-1, 0, 0, 0, 1, -1, 0, 0, -1),
                                    (-1, 0, 1, 0, -1, 1, 0, 0, 1),
                                    (0, -1, 0, -1, 0, 0, 0, 0, -1),
                                    (0, -1, 1, 1, 0, 0, 0, 0, 1),
                                    (0, 1, -1, 1, 0, -1, 0, 0, -1),
                                    (0, 1, 0, -1, 0, 1, 0, 0, 1),
                                    (1, 0, -1, 0, -1, 0, 0, 0, -1)]
                    else:
                        # borders 5, 6, 7
                        return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                                (-1, 0, 0, 0, 1, -1, 0, 0, -1),
                                (-1, 0, 1, 0, -1, 1, 0, 0, 1),
                                (1, 0, -1, 0, -1, 0, 0, 0, -1)]
            elif self._border(11):
                # borders 5, 11
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (-1, 0, 0, 0, 1, 0, 0, 0, -1),
                        (-1, 0, 1, 0, -1, 0, 0, 0, 1),
                        (1, 0, -1, 0, -1, 0, 0, 0, -1)]
            else:
                # borders 5
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (1, 0, -1, 0, -1, 0, 0, 0, -1)]

        if self._border(6):
            if self._border(12):
                if self._border(9):
                    # borders 6, 9, 12
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, 0, 0, 0, -1, 0, 0, -1, 1),
                            (-1, 0, 0, 0, -1, 1, 0, 0, 1),
                            (-1, 0, 0, 0, 0, -1, 0, -1, 0),
                            (-1, 0, 0, 0, 0, 1, 0, 1, 0),
                            (-1, 0, 0, 0, 1, -1, 0, 0, -1),
                            (-1, 0, 0, 0, 1, 0, 0, 1, -1),
                            (1, 0, 0, 0, -1, 0, 0, 0, -1),
                            (1, 0, 0, 0, -1, 1, 0, -1, 0),
                            (1, 0, 0, 0, 0, -1, 0, 1, -1),
                            (1, 0, 0, 0, 0, 1, 0, -1, 1),
                            (1, 0, 0, 0, 1, -1, 0, 1, 0)]
                else:
                    # borders 6, 12
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, 0, 0, 0, -1, 1, 0, 0, 1),
                            (-1, 0, 0, 0, 1, -1, 0, 0, -1),
                            (1, 0, 0, 0, -1, 0, 0, 0, -1)]
            else:
                # borders 6
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (-1, 0, 0, 0, 1, -1, 0, 0, -1)]

        if self._border(7):
            if self._border(8) and self._border(15):
                if self._border(16):
                    if self._border(9):
                        # borders 7, 8, 9, 15, 16
                        return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                                (-1, 0, 0, -1, 0, 1, -1, 1, 0),
                                (-1, 0, 0, 0, 0, -1, 0, -1, 0),
                                (-1, 0, 1, -1, 1, 0, -1, 0, 0),
                                (-1, 0, 1, 0, -1, 1, 0, 0, 1),
                                (-1, 1, 0, -1, 0, 0, -1, 0, 1),
                                (-1, 1, 0, 0, 1, 0, 0, 1, -1),
                                (0, -1, 0, -1, 0, 0, 0, 0, -1),
                                (0, -1, 0, 1, -1, 0, 0, -1, 1),
                                (0, -1, 1, 0, -1, 0, 1, -1, 0),
                                (0, -1, 1, 0, 0, 1, -1, 0, 1),
                                (0, 0, -1, 0, -1, 0, -1, 0, 0),
                                (0, 0, -1, 0, 1, -1, 1, 0, -1),
                                (0, 0, 1, -1, 0, 1, 0, -1, 1),
                                (0, 0, 1, 1, 0, 0, 0, 1, 0),
                                (0, 1, -1, -1, 1, 0, 0, 1, 0),
                                (0, 1, -1, 1, 0, -1, 0, 0, -1),
                                (0, 1, 0, 0, 0, 1, 1, 0, 0),
                                (0, 1, 0, 0, 1, -1, -1, 1, 0),
                                (1, -1, 0, 0, -1, 1, 0, -1, 0),
                                (1, -1, 0, 1, 0, -1, 1, 0, 0),
                                (1, 0, -1, 0, 0, -1, 0, 1, -1),
                                (1, 0, -1, 1, 0, 0, 1, -1, 0),
                                (1, 0, 0, 1, -1, 0, 1, 0, -1)]
                    else:
                        # borders 7, 8, 15, 16
                        return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                                (-1, 0, 0, -1, 0, 1, -1, 1, 0),
                                (-1, 0, 1, 0, -1, 1, 0, 0, 1),
                                (0, -1, 0, -1, 0, 0, 0, 0, -1),
                                (0, -1, 1, 0, -1, 0, 1, -1, 0),
                                (0, 1, -1, 1, 0, -1, 0, 0, -1),
                                (0, 1, 0, 0, 1, -1, -1, 1, 0),
                                (1, 0, -1, 1, 0, 0, 1, -1, 0)]
                else:
                    # borders 7, 8, 15
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, 0, 1, 0, -1, 1, 0, 0, 1),
                            (0, -1, 0, -1, 0, 0, 0, 0, -1),
                            (0, 1, -1, 1, 0, -1, 0, 0, -1)]
            elif self._border(9):
                # borders 7, 9
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (-1, 0, 0, 0, 0, -1, 0, -1, 0),
                        (-1, 0, 1, 0, -1, 1, 0, 0, 1),
                        (-1, 1, 0, 0, 1, 0, 0, 1, -1),
                        (1, -1, 0, 0, -1, 1, 0, -1, 0),
                        (1, 0, -1, 0, 0, -1, 0, 1, -1)]
            else:
                # borders 7
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (-1, 0, 1, 0, -1, 1, 0, 0, 1)]

        if self._border(8):
            if self._border(9):
                if self._border(10) and self._border(11) and self._border(12):
                    # borders 8, 9, 10, 11, 12
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, 0, 0, 0, -1, 0, 0, 0, 1),
                            (-1, 0, 0, 0, 0, -1, 0, -1, 0),
                            (-1, 0, 0, 0, 0, 1, 0, 1, 0),
                            (-1, 0, 0, 0, 1, 0, 0, 0, -1),
                            (0, -1, 0, -1, 0, 0, 0, 0, -1),
                            (0, -1, 0, 0, 0, -1, 1, 0, 0),
                            (0, -1, 0, 0, 0, 1, -1, 0, 0),
                            (0, -1, 0, 1, 0, 0, 0, 0, 1),
                            (0, 0, -1, -1, 0, 0, 0, 1, 0),
                            (0, 0, -1, 0, -1, 0, -1, 0, 0),
                            (0, 0, -1, 0, 1, 0, 1, 0, 0),
                            (0, 0, -1, 1, 0, 0, 0, -1, 0),
                            (0, 0, 1, -1, 0, 0, 0, -1, 0),
                            (0, 0, 1, 0, -1, 0, 1, 0, 0),
                            (0, 0, 1, 0, 1, 0, -1, 0, 0),
                            (0, 0, 1, 1, 0, 0, 0, 1, 0),
                            (0, 1, 0, -1, 0, 0, 0, 0, 1),
                            (0, 1, 0, 0, 0, -1, -1, 0, 0),
                            (0, 1, 0, 0, 0, 1, 1, 0, 0),
                            (0, 1, 0, 1, 0, 0, 0, 0, -1),
                            (1, 0, 0, 0, -1, 0, 0, 0, -1),
                            (1, 0, 0, 0, 0, -1, 0, 1, 0),
                            (1, 0, 0, 0, 0, 1, 0, -1, 0)]
                elif self._border(13) and self._border(14):
                    # borders 8, 9, 13, 14
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, -1, -1, 0, 0, 1, 0, 1, 0),
                            (-1, -1, -1, 0, 1, 0, 1, 0, 0),
                            (-1, -1, -1, 1, 0, 0, 0, 0, 1),
                            (-1, 0, 0, 0, -1, 0, 1, 1, 1),
                            (-1, 0, 0, 0, 0, -1, 0, -1, 0),
                            (-1, 0, 0, 1, 1, 1, 0, 0, -1),
                            (0, -1, 0, -1, 0, 0, 0, 0, -1),
                            (0, -1, 0, 0, 0, -1, 1, 1, 1),
                            (0, -1, 0, 1, 1, 1, -1, 0, 0),
                            (0, 0, -1, -1, 0, 0, 1, 1, 1),
                            (0, 0, -1, 0, -1, 0, -1, 0, 0),
                            (0, 0, -1, 1, 1, 1, 0, -1, 0),
                            (0, 0, 1, -1, -1, -1, 1, 0, 0),
                            (0, 0, 1, 0, 1, 0, -1, -1, -1),
                            (0, 0, 1, 1, 0, 0, 0, 1, 0),
                            (0, 1, 0, -1, -1, -1, 0, 0, 1),
                            (0, 1, 0, 0, 0, 1, 1, 0, 0),
                            (0, 1, 0, 1, 0, 0, -1, -1, -1),
                            (1, 0, 0, -1, -1, -1, 0, 1, 0),
                            (1, 0, 0, 0, 0, 1, -1, -1, -1),
                            (1, 1, 1, -1, 0, 0, 0, -1, 0),
                            (1, 1, 1, 0, -1, 0, 0, 0, -1),
                            (1, 1, 1, 0, 0, -1, -1, 0, 0)]
                else:
                    # borders 8, 9
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, 0, 0, 0, 0, -1, 0, -1, 0),
                            (0, -1, 0, -1, 0, 0, 0, 0, -1),
                            (0, 0, -1, 0, -1, 0, -1, 0, 0),
                            (0, 0, 1, 1, 0, 0, 0, 1, 0),
                            (0, 1, 0, 0, 0, 1, 1, 0, 0)]
            elif self._border(10):
                if self._border(11) and self._border(12):
                    # borders 8, 10, 11, 12
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, 0, 0, 0, -1, 0, 0, 0, 1),
                            (-1, 0, 0, 0, 1, 0, 0, 0, -1),
                            (0, -1, 0, -1, 0, 0, 0, 0, -1),
                            (0, -1, 0, 1, 0, 0, 0, 0, 1),
                            (0, 1, 0, -1, 0, 0, 0, 0, 1),
                            (0, 1, 0, 1, 0, 0, 0, 0, -1),
                            (1, 0, 0, 0, -1, 0, 0, 0, -1)]
                else:
                    # borders 8, 10
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, 0, 0, 0, -1, 0, 0, 0, 1),
                            (0, -1, 0, -1, 0, 0, 0, 0, -1),
                            (0, 1, 0, 1, 0, 0, 0, 0, -1)]
            elif self._border(14):
                # borders 8, 13, 14
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (-1, -1, -1, 1, 0, 0, 0, 0, 1),
                        (-1, 0, 0, 1, 1, 1, 0, 0, -1),
                        (0, -1, 0, -1, 0, 0, 0, 0, -1),
                        (0, 1, 0, -1, -1, -1, 0, 0, 1),
                        (1, 1, 1, 0, -1, 0, 0, 0, -1)]
            else:
                # borders 8
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (0, -1, 0, -1, 0, 0, 0, 0, -1)]

        if self._border(9):
            if self._border(12):
                if self._border(10) and self._border(11):
                    # borders 9, 10, 11, 12
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, 0, 0, 0, -1, 0, 0, 0, 1),
                            (-1, 0, 0, 0, 0, -1, 0, -1, 0),
                            (-1, 0, 0, 0, 0, 1, 0, 1, 0),
                            (-1, 0, 0, 0, 1, 0, 0, 0, -1),
                            (1, 0, 0, 0, -1, 0, 0, 0, -1),
                            (1, 0, 0, 0, 0, -1, 0, 1, 0),
                            (1, 0, 0, 0, 0, 1, 0, -1, 0)]
                else:
                    # borders 9, 12
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, 0, 0, 0, 0, -1, 0, -1, 0),
                            (-1, 0, 0, 0, 0, 1, 0, 1, 0),
                            (1, 0, 0, 0, -1, 0, 0, 0, -1)]
            elif self._border(14):
                if self._border(13):
                    # borders 9, 13, 14
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, -1, -1, 0, 0, 1, 0, 1, 0),
                            (-1, 0, 0, 0, 0, -1, 0, -1, 0),
                            (1, 1, 1, 0, -1, 0, 0, 0, -1)]
                else:
                    # borders 9, 14
                    return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                            (-1, -1, -1, 0, 0, 1, 0, 1, 0),
                            (-1, 0, 0, 0, 0, -1, 0, -1, 0),
                            (1, 1, 1, 0, -1, 0, 0, 0, -1)]
            elif self._border(15):
                # borders 9, 15, 16
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (-1, 0, 0, -1, 0, 1, -1, 1, 0),
                        (-1, 0, 0, 0, 0, -1, 0, -1, 0),
                        (0, -1, 1, 0, -1, 0, 1, -1, 0),
                        (0, -1, 1, 0, 0, 1, -1, 0, 1),
                        (0, 1, -1, -1, 1, 0, 0, 1, 0),
                        (0, 1, -1, 1, 0, -1, 0, 0, -1),
                        (1, 0, 0, 1, -1, 0, 1, 0, -1)]
            else:
                # borders 9
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (-1, 0, 0, 0, 0, -1, 0, -1, 0)]

        if self._border(10):
            if self._border(11) and self._border(12):
                # borders 10, 11, 12
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (-1, 0, 0, 0, -1, 0, 0, 0, 1),
                        (-1, 0, 0, 0, 1, 0, 0, 0, -1),
                        (1, 0, 0, 0, -1, 0, 0, 0, -1)]
            else:
                # borders 10
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (-1, 0, 0, 0, -1, 0, 0, 0, 1)]

        if self._border(11):
            # borders 11
            return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                    (-1, 0, 0, 0, 1, 0, 0, 0, -1)]

        if self._border(12):
            # border 12
            return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                    (1, 0, 0, 0, -1, 0, 0, 0, -1)]

        if self._border(13) and self._border(14):
            # border 13, 14
            return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                    (1, 1, 1, 0, -1, 0, 0, 0, -1)]

        if self._border(14):
            # border 14
            return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                    (1, 1, 1, 0, -1, 0, 0, 0, -1)]

        if self._border(15):
            if self._border(16):
                # borders 15, 16
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (-1, 0, 0, -1, 0, 1, -1, 1, 0),
                        (0, -1, 1, 0, -1, 0, 1, -1, 0),
                        (0, 1, -1, 1, 0, -1, 0, 0, -1)]
            else:
                # borders 15
                return [(1, 0, 0, 0, 1, 0, 0, 0, 1),
                        (0, 1, -1, 1, 0, -1, 0, 0, -1)]

        return [(1, 0, 0, 0, 1, 0, 0, 0, 1)]

    def _automorphisms_reduced_slow(self):
        """
        Return the automorphisms of the reduced ternary quadratic form.

        It searches over all 3x3 matrices with coefficients -1, 0, 1,
        determinant 1 and finite order, because Eisenstein reduced forms
        are Minkowski reduced. See Cassels.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 7, 0, 0, 0])
            sage: Q.is_eisenstein_reduced()
            True

            sage: # long time
            sage: auts = Q._automorphisms_reduced_slow()  # 3s on sage.math, 2014
            sage: len(auts)
            8
            sage: A = auts[randint(0,7)]
            sage: Q(A) == Q
            True
            sage: Q = TernaryQF([3, 4, 5, 3, 3, 2])
            sage: Q._automorphisms_reduced_slow()
            [
            [1 0 0]
            [0 1 0]
            [0 0 1]
            ]
        """
        from itertools import product
        if TernaryQF.possible_automorphisms is None:
            auts = (matrix(ZZ, 3, 3, m) for m in product([-1, 0, 1], repeat=9))
            auts = [m for m in auts if m.det() == 1]
            auts = [m for m in auts if m**2 in auts]
            auts = [m for m in auts if m**2 in auts]
            auts = [m for m in auts if m**2 in auts]
            TernaryQF.possible_automorphisms = auts
        return [m for m in TernaryQF.possible_automorphisms if self(m) == self]

    def automorphisms(self, slow=True):
        """
        Return a list with the automorphisms of the definite ternary quadratic form.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 7, 0, 0, 0])
            sage: auts = Q.automorphisms(); auts
            [
            [-1  0  0]  [-1  0  0]  [ 0 -1  0]  [ 0 -1  0]  [ 0  1  0]  [ 0  1  0]
            [ 0 -1  0]  [ 0  1  0]  [-1  0  0]  [ 1  0  0]  [-1  0  0]  [ 1  0  0]
            [ 0  0  1], [ 0  0 -1], [ 0  0 -1], [ 0  0  1], [ 0  0  1], [ 0  0 -1],
            [ 1  0  0]  [1 0 0]
            [ 0 -1  0]  [0 1 0]
            [ 0  0 -1], [0 0 1]
            ]
            sage: all(Q == Q(A) for A in auts)
            True
            sage: Q = TernaryQF([3, 4, 5, 3, 3, 2])
            sage: Q.automorphisms(slow=False)
            [
            [1 0 0]
            [0 1 0]
            [0 0 1]
            ]
            sage: Q = TernaryQF([4, 2, 4, 3, -4, -5])
            sage: auts = Q.automorphisms(slow=False)
            sage: auts
            [
            [1 0 0]  [ 2 -1 -1]
            [0 1 0]  [ 3 -2 -1]
            [0 0 1], [ 0  0 -1]
            ]
            sage: A = auts[1]
            sage: Q(A) == Q
            True
            sage: Qr, M_red = Q.reduced_form_eisenstein()
            sage: Qr
            Ternary quadratic form with integer coefficients:
            [1 2 3]
            [-1 0 -1]
            sage: Q(A*M_red) == Qr
            True
        """
        if not self.is_definite():
            raise ValueError("only implemented for definite forms")

        if self._automorphisms is not None:
            return self._automorphisms

        if self.is_positive_definite():
            if self.is_eisenstein_reduced():
                if slow:
                    self._automorphisms = self._automorphisms_reduced_slow()
                else:
                    auts = self._automorphisms_reduced_fast()
                    self._automorphisms = [matrix(ZZ, 3, A) for A in auts]
            else:
                Qr, M = self.reduced_form_eisenstein()
                auts = Qr.automorphisms(slow)
                M_inv = M.inverse()
                self._automorphisms = [M*m*M_inv for m in auts]
        else:
            self._automorphisms = (-self).automorphisms()
        return self._automorphisms

    def _number_of_automorphisms_reduced(self):
        """
        Return the number of automorphisms of the reduced definite ternary quadratic form.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 7, 0, 0, 0])
            sage: Q._number_of_automorphisms_reduced()
            8
            sage: len(Q.automorphisms(slow=False))
            8
            sage: Q = TernaryQF([3, 4, 5, 3, 3, 2])
            sage: Q._number_of_automorphisms_reduced()
            1
        """
        if self._border(1):
            if self._border(2):
                if self._border(14):
                    if self._border(9):
                        # borders 1, 2, 9, 14
                        return 8
                    else:
                        # borders 1, 2, 14
                        return 4
            else:
                # borders 1
                return 2

        if self._border(2):
            # borders 2
            return 2

        if self._border(3):
            # borders 3
            return 2

        if self._border(4):
            if self._border(10):
                if self._border(8):
                    # borders 4, 8, 10
                    return 12
                else:
                    # borders 4, 10
                    return 4
            else:
                # borders 4
                return 2

        if self._border(5):
            if self._border(6):
                if self._border(7):
                    if self._border(8):
                        if self._border(15):
                            # borders 5, 6, 7, 8, 15
                            return 8
                    else:
                        # borders 5, 6, 7
                        return 4
            elif self._border(11):
                # borders 5, 11
                return 4
            else:
                # borders 5
                return 2

        if self._border(6):
            if self._border(12):
                if self._border(9):
                    # borders 6, 9, 12
                    return 12
                else:
                    # borders 6, 12
                    return 4
            else:
                # borders 6
                return 2

        if self._border(7):
            if self._border(8) and self._border(15):
                if self._border(16):
                    if self._border(9):
                        # borders 7, 8, 9, 15, 16
                        return 24
                    else:
                        # borders 7, 8, 15, 16
                        return 8
                else:
                    # borders 7, 8, 15
                    return 4
            elif self._border(9):
                # borders 7, 9
                return 6
            else:
                # borders 7
                return 2

        if self._border(8):
            if self._border(9):
                if self._border(10) and self._border(11) and self._border(12):
                    # borders 8, 9, 10, 11, 12
                    return 24
                elif self._border(13) and self._border(14):
                    # borders 8, 9, 13, 14
                    return 24
                else:
                    # borders 8, 9
                    return 6
            elif self._border(10):
                if self._border(11) and self._border(12):
                    # borders 8, 10, 11, 12
                    return 8
                else:
                    # borders 8, 10
                    return 4
            elif self._border(14):
                # borders 8, 13, 14
                return 6
            else:
                # borders 8
                return 2

        if self._border(9):
            if self._border(12):
                if self._border(10) and self._border(11):
                    # borders 9, 10, 11, 12
                    return 8
                else:
                    # borders 9, 12
                    return 4
            elif self._border(14):
                if self._border(13):
                    # borders 9, 13, 14
                    return 4
                else:
                    # borders 9, 14
                    return 4
            elif self._border(15):
                # borders 9, 15, 16
                return 8
            else:
                # borders 9
                return 2

        if self._border(10):
            if self._border(11) and self._border(12):
                # borders 10, 11, 12
                return 4
            else:
                # borders 10
                return 2

        if self._border(11):
            # borders 11
            return 2

        if self._border(12):
            # border 12
            return 2

        if self._border(13) and self._border(14):
            # border 13, 14
            return 2

        if self._border(14):
            # border 14
            return 2

        if self._border(15):
            if self._border(16):
                # borders 15, 16
                return 4
            else:
                # borders 15
                return 2

        return 1

    def number_of_automorphisms(self, slow=True):
        """
        Return the number of automorphisms of the definite ternary quadratic form.

        EXAMPLES::

            sage: Q = TernaryQF([1, 1, 7, 0, 0, 0])
            sage: A = matrix(ZZ, 3, [0, 1, 0, -1, 5, 0, -8, -1, 1])
            sage: A.det()
            1
            sage: Q1 = Q(A); Q1
            Ternary quadratic form with integer coefficients:
            [449 33 7]
            [-14 -112 102]
            sage: Q1.number_of_automorphisms()
            8
            sage: Q = TernaryQF([-19, -7, -6, -12, 20, 23])
            sage: Q.is_negative_definite()
            True
            sage: Q.number_of_automorphisms(slow=False)
            24
        """
        if not self.is_definite():
            raise ValueError("only implemented for definite forms")

        if self._number_of_automorphisms is not None:
            return self._number_of_automorphisms

        if slow:
            self._number_of_automorphisms = len(self.automorphisms())
        else:
            if self.is_negative_definite():
                self._number_of_automorphisms = (-self).reduced_form_eisenstein(False)._number_of_automorphisms_reduced()
            else:
                self._number_of_automorphisms = self.reduced_form_eisenstein(False)._number_of_automorphisms_reduced()

        return self._number_of_automorphisms


def find_all_ternary_qf_by_level_disc(N, d):
    """
    Find the coefficients of all the reduced ternary quadratic forms given its discriminant `d` and level `N`.

    If `N|4d` and `d|N^2`, then it may be some forms with that discriminant and level.

    EXAMPLES::

        sage: find_all_ternary_qf_by_level_disc(44, 11)
        [Ternary quadratic form with integer coefficients:
        [1 1 3]
        [0 -1 0], Ternary quadratic form with integer coefficients:
        [1 1 4]
        [1 1 1]]
        sage: find_all_ternary_qf_by_level_disc(44, 11^2 * 16)
        [Ternary quadratic form with integer coefficients:
        [3 15 15]
        [-14 -2 -2], Ternary quadratic form with integer coefficients:
        [4 11 12]
        [0 -4 0]]
        sage: Q = TernaryQF([1, 1, 3, 0, -1, 0])
        sage: Q.is_eisenstein_reduced()
        True
        sage: Q.reciprocal_reduced()
        Ternary quadratic form with integer coefficients:
        [4 11 12]
        [0 -4 0]
        sage: find_all_ternary_qf_by_level_disc(44, 22)
        []
        sage: find_all_ternary_qf_by_level_disc(44, 33)
        Traceback (most recent call last):
        ...
        ValueError: There are no ternary forms of this level and discriminant
    """
    return [TernaryQF(qf) for qf in _find_all_ternary_qf_by_level_disc(N, d)]


def find_a_ternary_qf_by_level_disc(N, d):
    """
    Find a reduced ternary quadratic form given its discriminant `d` and level `N`.
    If `N|4d` and `d|N^2`, then it may be a form with that discriminant and level.

    EXAMPLES::

        sage: Q1 = find_a_ternary_qf_by_level_disc(44, 11); Q1
        Ternary quadratic form with integer coefficients:
        [1 1 3]
        [0 -1 0]
        sage: Q2 = find_a_ternary_qf_by_level_disc(44, 11^2 * 16)
        sage: Q2
        Ternary quadratic form with integer coefficients:
        [3 15 15]
        [-14 -2 -2]
        sage: Q1.is_eisenstein_reduced()
        True
        sage: Q1.level()
        44
        sage: Q1.disc()
        11
        sage: find_a_ternary_qf_by_level_disc(44, 22)
        sage: find_a_ternary_qf_by_level_disc(44, 33)
        Traceback (most recent call last):
        ...
        ValueError: There are no ternary forms of this level and discriminant
    """
    q = _find_a_ternary_qf_by_level_disc(N, d)
    if q is not None:
        return TernaryQF(q)



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.libs.gsl, known bug: windows (crash - https://github.com/passagemath/passagemath/issues/1892#issue-3726343166)
"""
GSL arrays
"""

from cysignals.memory cimport sig_malloc, sig_free

cdef class GSLDoubleArray:
    """
    EXAMPLES::

        sage: a = WaveletTransform(128,'daubechies',4)
        sage: for i in range(1, 11):
        ....:     a[i] = 1
        sage: a[:6:2]
        [0.0, 1.0, 1.0]
    """
    def __init__(self, size_t n, size_t stride=1, data=None):
        """
        EXAMPLES::

            sage: from sage.libs.gsl.array import GSLDoubleArray
            sage: a = GSLDoubleArray(10)
        """
        cdef int i

        self.n = n
        self.stride = stride
        self.data = <double *> sig_malloc(sizeof(double)*n)
        if data is not None:
            for i from 0 <= i < n:
                self.data[i] = data[i]
        else:
            for i from 0 <= i < n:
                self.data[i] = 0

    def __dealloc__(self):
        """
        EXAMPLES::

            sage: from sage.libs.gsl.array import GSLDoubleArray
            sage: a = GSLDoubleArray(10)
            sage: del a
        """
        sig_free(self.data)

    def __len__(self):
        """
        EXAMPLES::

            sage: from sage.libs.gsl.array import GSLDoubleArray
            sage: a = GSLDoubleArray(10)
            sage: len(a)
            10
        """
        return self.n

    def __repr__(self):
        """
        EXAMPLES::

            sage: from sage.libs.gsl.array import GSLDoubleArray
            sage: a = GSLDoubleArray(10)
            sage: a
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        """
        return str(list(self))

    def __setitem__(self, size_t i, x):
        """
        EXAMPLES::

            sage: from sage.libs.gsl.array import GSLDoubleArray
            sage: a = GSLDoubleArray(10)
            sage: a[5] = 3
            sage: a
            [0.0, 0.0, 0.0, 0.0, 0.0, 3.0, 0.0, 0.0, 0.0, 0.0]
        """
        # just set real for now
        if i < 0 or i >= self.n:
            raise IndexError
        self.data[i] = x

    def __getitem__(self, i):
        """
        EXAMPLES::

            sage: from sage.libs.gsl.array import GSLDoubleArray
            sage: a = GSLDoubleArray(10)
            sage: for i in range(10):
            ....:     a[i] = i
            sage: a[3:7]
            [3.0, 4.0, 5.0, 6.0]
        """
        if isinstance(i, slice):
            start, stop, step = i.indices(len(self))
            # TODO -- make this actually fast.
            return list(self)[start:stop:step]
        else:
            if i < 0 or i >= self.n:
                raise IndexError
            return self.data[i]



# sage_setup: distribution = sagemath-schemes
# sage.doctest: needs sage.libs.flint
r"""
Modular symbols `\{\alpha`, `\beta\}`

The ModularSymbol class represents a single modular symbol `X^i Y^{k-2-i} \{\alpha, \beta\}`.

AUTHOR:

- William Stein (2005, 2009)

TESTS::

    sage: s = ModularSymbols(11).2.modular_symbol_rep()[0][1]; s
    {-1/9, 0}
    sage: loads(dumps(s)) == s
    True
"""
# ****************************************************************************
#       Sage: Open Source Mathematical Software
#
#       Copyright (C) 2005, 2009 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
import sage.modular.cusps as cusps
from sage.modular.modsym.apply import apply_to_monomial
from sage.modular.modsym.manin_symbol import ManinSymbol
from sage.structure.sage_object import SageObject
import sage.structure.formal_sum as formal_sum
from sage.structure.richcmp import richcmp_method, richcmp
from sage.rings.integer_ring import ZZ
from sage.misc.latex import latex

_C = cusps.Cusps

X, Y = ZZ['X,Y'].gens()


@richcmp_method
class ModularSymbol(SageObject):
    r"""
    The modular symbol `X^i\cdot Y^{k-2-i}\cdot \{\alpha, \beta\}`.
    """
    def __init__(self, space, i, alpha, beta):
        """
        Initialise a modular symbol.

        INPUT:

        - ``space`` -- space of Manin symbols

        - ``i`` -- integer

        - ``alpha`` -- cusp

        - ``beta`` -- cusp

        EXAMPLES::

            sage: s = ModularSymbols(11).2.modular_symbol_rep()[0][1]; s
            {-1/9, 0}
            sage: type(s)
            <class 'sage.modular.modsym.modular_symbols.ModularSymbol'>
            sage: s = ModularSymbols(11,4).2.modular_symbol_rep()[0][1]; s
            X^2*{-1/7, 0}
        """
        self.__space = space
        self.__i = i
        self.__alpha = _C(alpha)
        self.__beta = _C(beta)

    def _repr_(self):
        """
        String representation of this modular symbol.

        EXAMPLES::

            sage: s = ModularSymbols(11,4).2.modular_symbol_rep()[0][1]; s
            X^2*{-1/7, 0}
            sage: s._repr_()
            'X^2*{-1/7, 0}'
            sage: s.rename('sym')
            sage: s
            sym
        """
        if self.weight() == 2:
            polypart = ''
        else:
            polypart = str(self.polynomial_part()) + '*'
        return "%s{%s, %s}" % (polypart, self.__alpha, self.__beta)

    def __getitem__(self, j):
        r"""
        Given a modular symbols `s = X^i Y^{k-2-i}\{\alpha, \beta\}`, ``s[0]`` is `\alpha`
        and ``s[1]`` is `\beta`.

        EXAMPLES::

            sage: s = ModularSymbols(11).2.modular_symbol_rep()[0][1]; s
            {-1/9, 0}
            sage: s[0]
            -1/9
            sage: s[1]
            0
            sage: s[2]
            Traceback (most recent call last):
            ...
            IndexError: list index out of range
        """
        return [self.__alpha, self.__beta][j]

    def _latex_(self):
        r"""
        Return Latex representation of this modular symbol.

        EXAMPLES::

            sage: s = ModularSymbols(11,4).2.modular_symbol_rep()[0][1]; s
            X^2*{-1/7, 0}
            sage: latex(s)                         # indirect doctest
            X^{2}\left\{\frac{-1}{7}, 0\right\}
        """
        if self.weight() == 2:
            polypart = ''
        else:
            polypart = latex(self.polynomial_part())
        return "%s\\left\\{%s, %s\\right\\}" % (polypart,
                                                latex(self.__alpha),
                                                latex(self.__beta))

    def __richcmp__(self, other, op):
        """
        Compare ``self`` to ``other``.

        EXAMPLES::

            sage: M = ModularSymbols(11)
            sage: s = M.2.modular_symbol_rep()[0][1]
            sage: t = M.0.modular_symbol_rep()[0][1]
            sage: s, t
            ({-1/9, 0}, {Infinity, 0})
            sage: s < t
            True
            sage: t > s
            True
            sage: s == s
            True
            sage: t == t
            True
        """
        if not isinstance(other, ModularSymbol):
            return NotImplemented
        return richcmp((self.__space, -self.__i, self.__alpha, self.__beta),
                       (other.__space,-other.__i,other.__alpha,other.__beta),
                       op)

    def __hash__(self):
        """
        EXAMPLES::

            sage: s = ModularSymbols(11).2.modular_symbol_rep()[0][1]
            sage: hash(s)  # random
            -7344656798833624820
        """
        return hash((self.__space, self.__i, self.__alpha, self.__beta))

    def space(self):
        """
        The list of Manin symbols to which this symbol belongs.

        EXAMPLES::

            sage: s = ModularSymbols(11).2.modular_symbol_rep()[0][1]
            sage: s.space()
            Manin Symbol List of weight 2 for Gamma0(11)
        """
        return self.__space

    def polynomial_part(self):
        r"""
        Return the polynomial part of this symbol, i.e. for a symbol of the
        form `X^i Y^{k-2-i}\{\alpha, \beta\}`, return `X^i Y^{k-2-i}`.

        EXAMPLES::

            sage: s = ModularSymbols(11).2.modular_symbol_rep()[0][1]
            sage: s.polynomial_part()
            1
            sage: s = ModularSymbols(1,28).0.modular_symbol_rep()[0][1]; s
            X^22*Y^4*{0, Infinity}
            sage: s.polynomial_part()
            X^22*Y^4
        """
        i = self.__i
        return X**i*Y**(self.weight()-2-i)

    def i(self):
        r"""
        For a symbol of the form `X^i Y^{k-2-i}\{\alpha, \beta\}`, return `i`.

        EXAMPLES::

            sage: s = ModularSymbols(11).2.modular_symbol_rep()[0][1]
            sage: s.i()
            0
            sage: s = ModularSymbols(1,28).0.modular_symbol_rep()[0][1]; s
            X^22*Y^4*{0, Infinity}
            sage: s.i()
            22
        """
        return self.__i

    def weight(self):
        r"""
        Return the weight of the modular symbols space to which this symbol
        belongs; i.e. for a symbol of the form `X^i Y^{k-2-i}\{\alpha,
        \beta\}`, return `k`.

        EXAMPLES::

            sage: s = ModularSymbols(1,28).0.modular_symbol_rep()[0][1]
            sage: s.weight()
            28
        """
        return self.__space.weight()

    def alpha(self):
        r"""
        For a symbol of the form `X^i Y^{k-2-i}\{\alpha, \beta\}`, return `\alpha`.

        EXAMPLES::

            sage: s = ModularSymbols(11,4).1.modular_symbol_rep()[0][1]; s
            X^2*{-1/6, 0}
            sage: s.alpha()
            -1/6
            sage: type(s.alpha())
            <class 'sage.modular.cusps.Cusp'>
        """
        return self.__alpha

    def beta(self):
        r"""
        For a symbol of the form `X^i Y^{k-2-i}\{\alpha, \beta\}`, return `\beta`.

        EXAMPLES::

            sage: s = ModularSymbols(11,4).1.modular_symbol_rep()[0][1]; s
            X^2*{-1/6, 0}
            sage: s.beta()
            0
            sage: type(s.beta())
            <class 'sage.modular.cusps.Cusp'>
        """
        return self.__beta

    def apply(self, g):
        r"""
        Act on this symbol by the element `g \in {\rm GL}_2(\QQ)`.

        INPUT:

        - ``g`` -- list ``[a,b,c,d]``, corresponding to the 2x2 matrix
          `\begin{pmatrix} a & b \\ c & d \end{pmatrix} \in {\rm GL}_2(\QQ)`

        OUTPUT:

        - ``FormalSum`` -- a formal sum `\sum_i c_i x_i`, where `c_i` are
          scalars and `x_i` are ModularSymbol objects, such that the sum
          `\sum_i c_i x_i` is the image of this symbol under the action of g.
          No reduction is performed modulo the relations that hold in
          self.space().

        The action of `g` on symbols is by

        .. MATH::

           P(X,Y)\{\alpha, \beta\} \mapsto  P(dX-bY, -cx+aY) \{g(\alpha), g(\beta)\}.

        Note that for us we have `P=X^i Y^{k-2-i}`, which simplifies computation
        of the polynomial part slightly.

        EXAMPLES::

            sage: s = ModularSymbols(11,2).1.modular_symbol_rep()[0][1]; s
            {-1/8, 0}
            sage: a = 1; b = 2; c = 3; d = 4; s.apply([a,b,c,d])
            {15/29, 1/2}
            sage: x = -1/8;  (a*x+b)/(c*x+d)
            15/29
            sage: x = 0;  (a*x+b)/(c*x+d)
            1/2
            sage: s = ModularSymbols(11,4).1.modular_symbol_rep()[0][1]; s
            X^2*{-1/6, 0}
            sage: s.apply([a,b,c,d])
            16*X^2*{11/21, 1/2} - 16*X*Y*{11/21, 1/2} + 4*Y^2*{11/21, 1/2}
            sage: P = s.polynomial_part()
            sage: X, Y = P.parent().gens()
            sage: P(d*X-b*Y, -c*X+a*Y)
            16*X^2 - 16*X*Y + 4*Y^2
            sage: x = -1/6; (a*x+b)/(c*x+d)
            11/21
            sage: x = 0; (a*x+b)/(c*x+d)
            1/2
            sage: type(s.apply([a,b,c,d]))
            <class 'sage.structure.formal_sum.FormalSum'>
        """
        space = self.__space
        i = self.__i
        k = space.weight()
        a, b, c, d = tuple(g)
        coeffs = apply_to_monomial(i, k - 2, d, -b, -c, a)
        g_alpha = self.__alpha.apply(g)
        g_beta = self.__beta.apply(g)
        return formal_sum.FormalSum([(coeffs[j], ModularSymbol(space, j, g_alpha, g_beta))
                                     for j in reversed(range(k-1)) if coeffs[j] != 0])

    def __manin_symbol_rep(self, alpha):
        """
        Return Manin symbol representation of X^i*Y^(k-2-i){0,alpha}.

        EXAMPLES::

            sage: s = ModularSymbols(11,2).1.modular_symbol_rep()[0][1]; s
            {-1/8, 0}
            sage: s.manin_symbol_rep()          # indirect doctest
            -(1,1) - (-8,1)
            sage: M = ModularSymbols(11,2)
            sage: s = M( (1,9) ); s
            (1,9)
            sage: t = s.modular_symbol_rep()[0][1].manin_symbol_rep(); t
            -(1,1) - (-9,1)
            sage: M(t)
            (1,9)
        """
        space = self.__space
        i = self.__i
        k = space.weight()
        v = [(0,1), (1,0)]
        if not alpha.is_infinity():
            cf = alpha._rational_().continued_fraction()
            v.extend((cf.p(k),cf.q(k)) for k in range(len(cf)))
        sign = 1
        z = formal_sum.FormalSum(0)
        for j in range(1,len(v)):
            c = sign*v[j][1]
            d = v[j-1][1]
            coeffs = apply_to_monomial(i, k-2, sign*v[j][0], v[j-1][0],
                                       sign*v[j][1], v[j-1][1])
            w = [(coeffs[j], ManinSymbol(space, (j, c, d)))
                 for j in range(k-1) if coeffs[j] != 0]
            z += formal_sum.FormalSum(w)
            sign *= -1
        return z

    def manin_symbol_rep(self):
        """
        Return a representation of ``self`` as a formal sum of Manin symbols.

        The result is not cached.

        EXAMPLES::

            sage: M = ModularSymbols(11,4)
            sage: s = M.1.modular_symbol_rep()[0][1]; s
            X^2*{-1/6, 0}
            sage: s.manin_symbol_rep()
            -2*[X*Y,(-1,0)] - [X^2,(-1,0)] - [Y^2,(1,1)] - [X^2,(-6,1)]
            sage: M(s.manin_symbol_rep()) == M([2,-1/6,0])
            True
        """
        alpha = self.__alpha
        beta = self.__beta
        return -1*self.__manin_symbol_rep(alpha) + self.__manin_symbol_rep(beta)



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/nmod_mat.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void nmod_mat_init(nmod_mat_t mat, slong rows, slong cols, mp_limb_t n) noexcept
    void nmod_mat_init_set(nmod_mat_t mat, const nmod_mat_t src) noexcept
    void nmod_mat_clear(nmod_mat_t mat) noexcept
    void nmod_mat_set(nmod_mat_t mat, const nmod_mat_t src) noexcept
    void nmod_mat_swap(nmod_mat_t mat1, nmod_mat_t mat2) noexcept
    void nmod_mat_swap_entrywise(nmod_mat_t mat1, nmod_mat_t mat2) noexcept
    mp_limb_t nmod_mat_get_entry(const nmod_mat_t mat, slong i, slong j) noexcept
    mp_limb_t * nmod_mat_entry_ptr(const nmod_mat_t mat, slong i, slong j) noexcept
    void nmod_mat_set_entry(nmod_mat_t mat, slong i, slong j, mp_limb_t x) noexcept
    slong nmod_mat_nrows(const nmod_mat_t mat) noexcept
    slong nmod_mat_ncols(const nmod_mat_t mat) noexcept
    void nmod_mat_zero(nmod_mat_t mat) noexcept
    bint nmod_mat_is_zero(const nmod_mat_t mat) noexcept
    void nmod_mat_window_init(nmod_mat_t window, const nmod_mat_t mat, slong r1, slong c1, slong r2, slong c2) noexcept
    void nmod_mat_window_clear(nmod_mat_t window) noexcept
    void nmod_mat_concat_vertical(nmod_mat_t res, const nmod_mat_t mat1, const nmod_mat_t mat2) noexcept
    void nmod_mat_concat_horizontal(nmod_mat_t res, const nmod_mat_t mat1, const nmod_mat_t mat2) noexcept
    void nmod_mat_print_pretty(const nmod_mat_t mat) noexcept
    int nmod_mat_fprint_pretty(FILE * file, const nmod_mat_t mat) noexcept
    int nmod_mat_print(const nmod_mat_t mat) noexcept
    int nmod_mat_fprint(FILE * f, const nmod_mat_t mat) noexcept
    void nmod_mat_randtest(nmod_mat_t mat, flint_rand_t state) noexcept
    void nmod_mat_randfull(nmod_mat_t mat, flint_rand_t state) noexcept
    int nmod_mat_randpermdiag(nmod_mat_t mat, flint_rand_t state, mp_srcptr diag, slong n) noexcept
    void nmod_mat_randrank(nmod_mat_t mat, flint_rand_t state, slong rank) noexcept
    void nmod_mat_randops(nmod_mat_t mat, slong count, flint_rand_t state) noexcept
    void nmod_mat_randtril(nmod_mat_t mat, flint_rand_t state, int unit) noexcept
    void nmod_mat_randtriu(nmod_mat_t mat, flint_rand_t state, int unit) noexcept
    bint nmod_mat_equal(const nmod_mat_t mat1, const nmod_mat_t mat2) noexcept
    bint nmod_mat_is_zero_row(const nmod_mat_t mat, slong i) noexcept
    void nmod_mat_transpose(nmod_mat_t B, const nmod_mat_t A) noexcept
    void nmod_mat_swap_rows(nmod_mat_t mat, slong * perm, slong r, slong s) noexcept
    void nmod_mat_swap_cols(nmod_mat_t mat, slong * perm, slong r, slong s) noexcept
    void nmod_mat_invert_rows(nmod_mat_t mat, slong * perm) noexcept
    void nmod_mat_invert_cols(nmod_mat_t mat, slong * perm) noexcept
    void nmod_mat_permute_rows(nmod_mat_t mat, const slong * perm_act, slong * perm_store) noexcept
    void nmod_mat_add(nmod_mat_t C, const nmod_mat_t A, const nmod_mat_t B) noexcept
    void nmod_mat_sub(nmod_mat_t C, const nmod_mat_t A, const nmod_mat_t B) noexcept
    void nmod_mat_neg(nmod_mat_t A, const nmod_mat_t B) noexcept
    void nmod_mat_scalar_mul(nmod_mat_t B, const nmod_mat_t A, mp_limb_t c) noexcept
    void nmod_mat_scalar_addmul_ui(nmod_mat_t dest, const nmod_mat_t X, const nmod_mat_t Y, const mp_limb_t b) noexcept
    void nmod_mat_scalar_mul_fmpz(nmod_mat_t res, const nmod_mat_t M, const fmpz_t c) noexcept
    void nmod_mat_mul(nmod_mat_t C, const nmod_mat_t A, const nmod_mat_t B) noexcept
    void _nmod_mat_mul_classical_op(nmod_mat_t D, const nmod_mat_t C, const nmod_mat_t A, const nmod_mat_t B, int op) noexcept
    void nmod_mat_mul_classical(nmod_mat_t C, const nmod_mat_t A, const nmod_mat_t B) noexcept
    void _nmod_mat_mul_classical_threaded_pool_op(nmod_mat_t D, const nmod_mat_t C, const nmod_mat_t A, const nmod_mat_t B, int op, thread_pool_handle * threads, slong num_threads) noexcept
    void _nmod_mat_mul_classical_threaded_op(nmod_mat_t D, const nmod_mat_t C, const nmod_mat_t A, const nmod_mat_t B, int op) noexcept
    void nmod_mat_mul_classical_threaded(nmod_mat_t C, const nmod_mat_t A, const nmod_mat_t B) noexcept
    void nmod_mat_mul_strassen(nmod_mat_t C, const nmod_mat_t A, const nmod_mat_t B) noexcept
    int nmod_mat_mul_blas(nmod_mat_t C, const nmod_mat_t A, const nmod_mat_t B) noexcept
    void nmod_mat_addmul(nmod_mat_t D, const nmod_mat_t C, const nmod_mat_t A, const nmod_mat_t B) noexcept
    void nmod_mat_submul(nmod_mat_t D, const nmod_mat_t C, const nmod_mat_t A, const nmod_mat_t B) noexcept
    void nmod_mat_mul_nmod_vec(mp_limb_t * c, const nmod_mat_t A, const mp_limb_t * b, slong blen) noexcept
    void nmod_mat_mul_nmod_vec_ptr(mp_limb_t * const * c, const nmod_mat_t A, const mp_limb_t * const * b, slong blen) noexcept
    void nmod_mat_nmod_vec_mul(mp_limb_t * c, const mp_limb_t * a, slong alen, const nmod_mat_t B) noexcept
    void nmod_mat_nmod_vec_mul_ptr(mp_limb_t * const * c, const mp_limb_t * const * a, slong alen, const nmod_mat_t B) noexcept
    void _nmod_mat_pow(nmod_mat_t dest, const nmod_mat_t mat, ulong pow) noexcept
    void nmod_mat_pow(nmod_mat_t dest, const nmod_mat_t mat, ulong pow) noexcept
    mp_limb_t nmod_mat_trace(const nmod_mat_t mat) noexcept
    mp_limb_t nmod_mat_det_howell(const nmod_mat_t A) noexcept
    mp_limb_t nmod_mat_det(const nmod_mat_t A) noexcept
    slong nmod_mat_rank(const nmod_mat_t A) noexcept
    int nmod_mat_inv(nmod_mat_t B, const nmod_mat_t A) noexcept
    void nmod_mat_solve_tril(nmod_mat_t X, const nmod_mat_t L, const nmod_mat_t B, int unit) noexcept
    void nmod_mat_solve_tril_classical(nmod_mat_t X, const nmod_mat_t L, const nmod_mat_t B, int unit) noexcept
    void nmod_mat_solve_tril_recursive(nmod_mat_t X, const nmod_mat_t L, const nmod_mat_t B, int unit) noexcept
    void nmod_mat_solve_triu(nmod_mat_t X, const nmod_mat_t U, const nmod_mat_t B, int unit) noexcept
    void nmod_mat_solve_triu_classical(nmod_mat_t X, const nmod_mat_t U, const nmod_mat_t B, int unit) noexcept
    void nmod_mat_solve_triu_recursive(nmod_mat_t X, const nmod_mat_t U, const nmod_mat_t B, int unit) noexcept
    int nmod_mat_solve(nmod_mat_t X, const nmod_mat_t A, const nmod_mat_t B) noexcept
    int nmod_mat_can_solve_inner(slong * rank, slong * perm, slong * pivots, nmod_mat_t X, const nmod_mat_t A, const nmod_mat_t B) noexcept
    int nmod_mat_can_solve(nmod_mat_t X, const nmod_mat_t A, const nmod_mat_t B) noexcept
    int nmod_mat_solve_vec(mp_ptr x, const nmod_mat_t A, mp_srcptr b) noexcept
    slong nmod_mat_lu(slong * P, nmod_mat_t A, int rank_check) noexcept
    slong nmod_mat_lu_classical(slong * P, nmod_mat_t A, int rank_check) noexcept
    slong nmod_mat_lu_classical_delayed(slong * P, nmod_mat_t A, int rank_check) noexcept
    slong nmod_mat_lu_recursive(slong * P, nmod_mat_t A, int rank_check) noexcept
    slong nmod_mat_rref(nmod_mat_t A) noexcept
    slong nmod_mat_reduce_row(nmod_mat_t A, slong * P, slong * L, slong n) noexcept
    slong nmod_mat_nullspace(nmod_mat_t X, const nmod_mat_t A) noexcept
    void nmod_mat_similarity(nmod_mat_t M, slong r, ulong d) noexcept
    void nmod_mat_charpoly_berkowitz(nmod_poly_t p, const nmod_mat_t M) noexcept
    void nmod_mat_charpoly_danilevsky(nmod_poly_t p, const nmod_mat_t M) noexcept
    void nmod_mat_charpoly(nmod_poly_t p, const nmod_mat_t M) noexcept
    void nmod_mat_minpoly(nmod_poly_t p, const nmod_mat_t M) noexcept
    void nmod_mat_strong_echelon_form(nmod_mat_t A) noexcept
    slong nmod_mat_howell_form(nmod_mat_t A) noexcept



# sage_setup: distribution = sagemath-modules
#*****************************************************************************
#       Copyright (C) 2006 - 2011 Robert L. Miller <rlmillster@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.groups.perm_gps.partn_ref.data_structures cimport *

from sage.groups.perm_gps.partn_ref.automorphism_group_canonical_label cimport (
    get_aut_gp_and_can_lab, aut_gp_and_can_lab, agcl_work_space,
    allocate_agcl_output, deallocate_agcl_output,
    allocate_agcl_work_space, deallocate_agcl_work_space)

cdef class MatrixStruct:
    cdef list symbol_structs
    cdef object matrix
    cdef int degree
    cdef int nwords
    cdef list symbols
    cdef int nsymbols
    cdef PartitionStack *temp_col_ps
    cdef aut_gp_and_can_lab *output

cdef int refine_matrix(PartitionStack *, void *, int *, int) noexcept
cdef int compare_matrices(int *, int *, void *, void *, int) noexcept
cdef bint all_matrix_children_are_equivalent(PartitionStack *, void *) noexcept



# sage_setup: distribution = sagemath-pari
from cpython cimport array

cimport sage.structure.element
from sage.libs.gmp.types cimport mpz_t, mpq_t
from sage.structure.element cimport Element, RingElement
from sage.rings.padics.local_generic_element cimport LocalGenericElement
from sage.rings.padics.pow_computer cimport PowComputer_class
from sage.rings.integer cimport Integer
from sage.rings.rational cimport Rational

cpdef gauss_table(long long p, int f, int prec, bint use_longs)

cdef class pAdicGenericElement(LocalGenericElement):
    cdef long valuation_c(self) noexcept
    cpdef val_unit(self)

    cdef int _set_from_Integer(self, Integer x, absprec, relprec) except -1
    cdef int _set_from_mpz(self, mpz_t x) except -1
    cdef int _set_from_mpz_rel(self, mpz_t x, long relprec) except -1
    cdef int _set_from_mpz_abs(self, mpz_t value, long absprec) except -1
    cdef int _set_from_mpz_both(self, mpz_t x, long absprec, long relprec) except -1

    cdef int _set_from_Rational(self, Rational x, absprec, relprec) except -1
    cdef int _set_from_mpq(self, mpq_t x) except -1
    cdef int _set_from_mpq_rel(self, mpq_t x, long relprec) except -1
    cdef int _set_from_mpq_abs(self, mpq_t value, long absprec) except -1
    cdef int _set_from_mpq_both(self, mpq_t x, long absprec, long relprec) except -1

    cdef int _pshift_self(self, long shift) except -1

    cdef int _cmp_units(left, pAdicGenericElement right) except -2

    cdef int _set_inexact_zero(self, long absprec) except -1
    cdef int _set_exact_zero(self) except -1

    cpdef bint _is_exact_zero(self) except -1
    cpdef bint _is_inexact_zero(self) except -1
    cpdef bint _is_zero_rep(self) except -1

    cdef bint _set_prec_abs(self, long absprec) except -1
    cdef bint _set_prec_rel(self, long relprec) except -1
    cdef bint _set_prec_both(self, long absprec, long relprec) except -1

    cpdef abs(self, prec=*)
    cpdef _mod_(self, right)
    cpdef _floordiv_(self, right)
    cpdef bint _is_base_elt(self, p) except -1
